(* Copy, pack and unpack files, create zip archives
   optional with encryption

    J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   Acknowledgements:
     ZLibEx (Vers. 1.2.3) and ZLib routines from http://www.base2ti.com
     AES functions from http://fp.gladman.plus.com/index.htm

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Sept. 2005
   Vers. 2 - Oct. 2005   : Zip64 extensions
   Vers. 3 - Dec. 2005   : AE-2 encryption added (see AesLib)
   Vers. 4 - Apr. 2006   : Verify option added
   Vers. 4.1 - Jul. 2006 : some minor changes and corrections
   Vers. 4.2 - Aug. 2006 : new error: path too long, TZip.Destroy fixed (path>255)
   Vers. 4.3 - Sep. 2006 : exception in verify routines fixed
   Vers. 4.4 - Apr. 2007 : Warnings on setting attributes after copy
                           New error messages
   Vers. 4.5 - May  2007 : Verify on zip archives
   Vers. 4.6 - Jul. 2007 : allow setting of encryption depth on Zip files
   Vers. 4.7 - Oct. 2007 : alternate copy using SHFileOperation
                           without compression and encryption
   Vers. 4.8 - Jan. 2008 : full timestamp copy
   Vers. 5   - Apr. 2009 : Read zip files
   Vers. 5.1 - Jul. 2009 : Copy file storage
   *)

unit ExtFileTools;

interface

uses Windows, Classes, SysUtils, ZLibex, AesLib, FtGlobal, FileUtils;

const
  GzExt = '.gz';      // default extension for GZip
  (* Format of encrypted gz-files:
     Standard GZIP header,  bit 5 is set in flag byte
     Filename
     Encrypted data
     CRC32 set to 0
     Original size mod 2^32
  *)
  CrExt = '.enc';      // default extension for not compressed encrypted files
  defBufferSize = 256*1024;

  // ZLib parameters
  defWindowBits = -15;  // for GZip format
  defMemLevel = 8;

  // gz signature
  GzSignatur = $8b1f;
  Z_DEFLATE = 8; 
  Z_LZMA = 14;   // LZMA
  
  // gzip flag byte
  gfASCII      = $01; { bit 0 set: file probably ascii text }
  gfHeaderCrc  = $02; { bit 1 set: header CRC present }
  gfExtraField = $04; { bit 2 set: extra field present }
  gfFileName   = $08; { bit 3 set: original file name present }
  gfComment    = $10; { bit 4 set: file comment present }
  gfEncrypted  = $20; { bit 5 set: AE-2 encryption used for data after compressing }
  gfReserved   = $C0; { bits 6..7: reserved }

  // PkZip signatures
  PkLocalSignatur = $04034b50;
  PkDirSignatur = $02014b50;
  PkEndSignatur = $06054b50;
  Pk64RecSignatur = $06064b50;
  Pk64LocSignatur = $07064b50;

  // PkZip Constants
  VersDef = $14;
  VersZip64 = $2D;
  VersDos = 0;
  zipEncryptFlag = 1;     // File is encrypted
  zipMaxDeflFlag = 2;     // Deflating - Maximum (-exx/-ex) compression option was used.
  zipNoComp      = 0;     // File is stored
  zipDeflate     = 8;     // File is deflated
  zipLzma        = 14;    // Use LZMA compression
  zipEncrypted   = 99;    // Compression methode uses encryption
  ExtraZip64 = 1;         // Extra field tag for Zip64
  ExtraCrypt = $9901;     // Extra field tag for encryption
  VendorID   = $4541;     // 'AE' - see: http://www.winzip.de/aes_info.htm
  VendorVersAE1 = 1;
  VendorVersAE2 = 2;
  AE1Size       = 20;     // see: http://www.winzip.de/aes_info.htm#crc-faq
  AE2Mode128 = 1;
  AE2Mode192 = 2;
  AE2Mode256 = 3;

  tmpGzipName = 'gzip.tmp';
  tmpEncName = 'encrypt.tmp';
  UMaxMemSize = 10*1024*1024;  // max. size of temp. memory stream (uncompressed)
  CMaxMemSize = 1024*1024;  // max. size of temp. memory stream (compressed)
  MaxCardinal = $FFFFFFFF;    // max. value for cardinals (4 bytes)

type
  TGzHeader = packed record
    Signature : word;
    Methode,
    Flags     : byte;
    TimeStamp : cardinal;
    XFlags,
    OpSys     : byte;
    end;

  TGzFileInfo = record
    Filename  : string;          // original filename
    UnixTime  : cardinal;        // original timestamp as Unix time
    Attr      : word;            // original file attributes
    USize     : cardinal;        // original size
    Encrypted : boolean;         // file is encrypted
    end;

  TPKExtraLocalField = packed record
    Tag,ESize   : word;
    Buffer      : array [0..3] of int64;
    end;

  TPKExtraEncField = packed record
    Tag,ESize,
    Version,Vendor  : word;
    Mode            : byte;
    Method          : word;
    end;

  TPKLocalHeader = packed record
    Signatur    : cardinal;
    Vers,Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength : word;
    end;

  TPKDirHeader = packed record
    Signatur    : cardinal;
    VersMade,
    VersExtr,
    Flag,
    Method      : word;
    FTimeStamp,
    CRC,
    CSize,USize : cardinal;
    FNLength,
    ExtraLength,
    CommLength,
    DiskStart,
    IntAttr     : word;
    ExtAttr,
    Offset      : cardinal;
    end;

  TPKEndHeader = packed record
    Signatur    : cardinal;
    ThisDisk,
    StartDisk,
    ThisEntries,
    TotalEntries : word;
    DirSize,
    Offset      : cardinal;
    CommLength  : word;
    end;

  TPKZip64CdRecord = packed record
    Signatur    : cardinal;
    Size        : int64;
    VersMade,
    VersExtr    : word;
    ThisDisk,
    StartDisk   : cardinal;
    ThisEntries,
    TotalEntries,
    DirSize,
    Offset      : int64;
    end;

  TPKZip64CdLocator = packed record
    Signatur    : cardinal;
    StartDisk   : cardinal;
    Offset      : int64;
    TotalDisk   : cardinal;
    end;

  TZip64ExtraField = packed record
    Tag,ESize   : word;
    end;

  TZipFileInfo = record
    Filename        : string;
    Flag,Method     : word;
    TimeStamp,Crc   : cardinal;
    CSize,USize     : int64;
    end;

  TPkFileRef = class(TObject)
    Offset  : int64;
    constructor Create (Ofs : int64);
    end;

  EZipError = class(EInOutError)
  public
    ZipError : integer;
    constructor Create (const ErrString : string; ErrNR : integer);
    end;

  TPkHeader = class(TObject)
    FName           : string;  // Original filename (OEM)
    ErrCode,
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethode     : word;
    Offset,
    CSize,USize     : int64;
    IsDir,IsAE2,
    Zip64           : boolean;
    constructor Create (const Fn : string; Ec,Ts,Chk : cardinal; AA,AF,AM : word;
                        Ofs,CSz,USz : int64; AE2 : boolean);
    end;

const
  // Extra field for AES encryption (see: http://www.winzip.de/aes_info.htm)
  defPkExtraCrypt : TPKExtraEncField = (Tag : ExtraCrypt; ESize : 7; Version : VendorVersAE2;
                       Vendor : VendorID; Mode : AE2Mode128; Method : zipDeflate);

type
  TCopyOption = (coCompress,coVerify,coCreationTime,coStorage);
  TCopyOptions = set of TCopyOption;

  TShellCopyMode = (scmCopy,scmQuiet);
  TShellCopyModes = set of TShellCopyMode;

  // base class for all copy functions (compressed and crypted)
  TCopyThread = class (TThread)
  private
    FError           : integer;
    FOnProgress      : TProgressEvent;
    FCheckTime       : boolean;           // check for illegal filetime
    FShellCopy       : TShellCopyModes;   // use shell copy function, overrides compression and encryption
    function GetDone : boolean;
    procedure SetProgress (ACallBack : TProgressEvent);
  protected
    FBuffer          : array of byte;
    SourceName,
    DestName         : string;
    sSource,sDest    : TFileStream;
    FBufSize         : cardinal;
    FUserBreak,
    FCopyCrTm,
    FCopyStorage,
    FVerify          : boolean;
    FErrFlag         : integer;
    FCompLevel       : TZCompressionLevel;
    FCompType        : TZStrategy;
    FCrc             : cardinal;
    UTotal,CTotal    : int64;
    FCrypting        : TEncryption;
    FAction          : TFileAction;
    FBytesRead,
    FBytesWritten,
    FCount           : int64;
    function CopyStream (fSource,fDest : TStream; SLength : int64) : integer;
    function GzCompress (fSource,fDest : TStream) : integer;
    function GzUnCompress (fSource,fDest : TStream) : integer; overload;
    function GzUnCompress (fSource,fDest : TStream; SLength : int64) : integer; overload;
    function EncryptFile : integer;
    function DecryptFile : integer;
    procedure EndThread (Error : integer); virtual;
    procedure UpdateProgress;
    procedure DoProgress (AAction : TFileAction; ACount : int64); virtual;
    function OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
    procedure ReleaseStreams;
    function CloseFiles (FTime : TFileTimestamps; Attr : word) : integer;
    function CopyFile : integer;
    function ShCopyFile : integer;
    function VerifyFile : integer;
    function VerifyEncFile : integer;
    function CopyFileStorage : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest. file
    constructor Create (const ASourceName,ADestName : string;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    destructor Destroy; override;
    procedure CancelThread; virtual;
    property ErrorCode : integer read FError;
    property BytesRead : int64 read FBytesRead;
    property BytesWritten : int64 read FBytesWritten;
    property CheckTime : boolean read FCheckTime write FCheckTime;
    property CopyCreationTime : boolean read FCopyCrTm write FCopyCrTm;
    property ShellCopy : TShellCopyModes read FShellCopy write FShellCopy;
    property Done  : boolean read GetDone;
    property OnProgress : TProgressEvent read FOnProgress write SetProgress;
    end;

  // Copy file with opt. encryption
  TEncryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    end;

  // Decrypt file
  TDecryptThread = class (TCopyThread)
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // AEncryption = Encrypting object (see AesLib)
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    end;

  // gzip compatible compression
  TGZipThread = class (TCopyThread)
  protected
    function GzWriteHeader (fDest : TStream; fName : string;
                            fMethode,fTime : cardinal; fEnc : boolean) : integer;
    function GzWriteTrailer (fDest : TStream; ACrc,ASize : cardinal) : integer;
    function CompressFile : integer;
    function VerifyGzFile : integer;
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = gz-file,
  // AEncryption = encrypting object (see AesLib), set to nil for plain compressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    property CompressionLevel : TZCompressionLevel read FCompLevel write FCompLevel;
    property CompressionType : TZStrategy read FCompType write FCompType;
    property Crc : cardinal read FCrc;
    end;

  // copy file with opt. compression and/or encryption
  TWinCopyThread = class (TGZipThread)
  private
    FCompress  : boolean;
  protected
    procedure Execute; override;
  public
  // ASourceName  = source file,  ADestName = dest file,
  // ACompress    = use gzip compression
  // AVerify      = verify after operation
  // AEncryption  = encrypting object (see AesLib), set to nil for no encryption
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ACompress,AVerify,ASuspend : Boolean; ABufSize : integer;
                        APriority : TThreadPriority);
    property CopyStorage : boolean read FCopyStorage write FCopyStorage;
    end;

  // gzip compatible uncompression
  TGUnZipThread = class (TCopyThread)
  protected
    function GzReadHeader (fSource : TStream; var fName : string;
                           var fTime : cardinal; var fEnc  : boolean) : integer;
    function GzReadTrailer (fSource : TStream) : integer;
    function UnCompressFile : integer;
    procedure Execute; override;
  public
  // ASourceName  = gz-file,  ADestFolder = destination folder
  // ADecryption = Decrypting object (see AesLib), set to nil for plain uncompressing
    constructor Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
    property Crc : cardinal read FCrc;
    end;

  // Zip compatible compression with optional encryption
  TZipWriteThread = class (TGZipThread)
  protected
    function AddToZip : integer;
    procedure Execute; override;
  public
    constructor Create (const AFileName : string; AEncryption : TEncryption;
                        ZStream : TStream; ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent);
    end;

  // Zip compatible decompression with optional encryption for verification
  TZipVerifyThread = class (TGZipThread)
  private
    FLength : int64;
  protected
    function VerifyZippedFile : integer;
    procedure Execute; override;
  public
    constructor Create (AEncryption : TEncryption; ACrc : cardinal;
                        ZStream : TStream; ALength : int64; ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent);
    end;

  // Zip compatible decompression with optional encryption
  TZipReadThread = class (TGUnZipThread)
  protected
    function ReadZipLocalHeader(var ZipInfo : TZipFileInfo) : boolean;
    function ReadFromZip : integer;
    procedure Execute; override;
  public
    constructor Create (const AFileName : string; AEncryption : TEncryption;
                        ZStream : TStream; ASuspend : Boolean;
                        ABufSize : integer; APriority : TThreadPriority;
                        AOnProgress : TProgressEvent);
    end;

  TZip = class (TFileStream)
  private
    FBufferSize : integer;
    FEncrypt,
    FErrorFlag  : boolean;
    FEncryption : TEncryption;
    PkExtraCrypt : TPKExtraEncField;
    FOnProgress : TProgressEvent;
    procedure SetProgress (ACallBack : TProgressEvent);
  protected
    FPriority   : TThreadPriority;
    PBase,ZName : string;
    FileList    : TStringList;
  public
    constructor Create (const ZipName,Password : string; AKeyMode,ABufSize,AMode : integer);
    destructor Destroy; override;
    procedure InitEncryption (Password : string; AKeyMode : integer);
    property OnProgress : TProgressEvent read FOnProgress write SetProgress;
    property ErrorFlag : boolean read FErrorFlag write FErrorFlag;
    property ThreadPriority : TThreadPriority read FPriority write FPriority;
    end;

  TZipRead = class (TZip)
  private
    ZipThread   : TGUnZipThread;
  public
    constructor Create (const ZipName,Password : string; AKeyMode,ABufSize : integer);
    destructor Destroy; override;
    property Directory : TStringList read FileList;
    procedure ReadDirectory;
    function Filename(Index : integer) :string;
    function IsEncrypted(Index : integer) : boolean;
    procedure ExtractFromZip (Offset : int64; TimeStamp,Attr : integer; const DestFilename : string);
    procedure ExtractFileFromZip (Index : integer; const DestFilename : string);
    procedure Cancel;
    end;

  TZipWrite = class (TZip)
  private
    FOnVerifyProgress : TProgressEvent;
    ZipThread   : TGZipThread;
    FBytesRead,
    FBytesWritten : int64;
    function VerifyFile (PkHeader : TPkHeader) : boolean;
  public
    constructor Create (const ZipName,BasicDirectory,Password : string; AKeyMode,ABufSize : integer);
    destructor Destroy; override;
    procedure AddDir (DirName : string);
    procedure RemLastDir;
    procedure AddFile (SourceName : string);
    procedure WriteDirectory;
    procedure Verify;
    procedure Suspend;
    procedure Resume;
    property BytesRead : int64 read FBytesRead;
    property BytesWritten : int64 read FBytesWritten;
    property OnVerifyProgress : TProgressEvent read FOnVerifyProgress write FOnVerifyProgress;
    procedure Cancel;
    end;

// retrieve gz file info
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
function GzFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;

// check file extension for gz or enc
function IsGzFile (const Filename : string) : boolean;
function IsEncFile (const Filename : string) : boolean;

implementation

uses Forms, FilePropUtils, WinApiUtils, WinShell, HTTPApp;

const
  MaxSize = $FF000000;     // max. file size for standard zip

{ ---------------------------------------------------------------- }
type
  TCvTable = array [0..127] of char;
const
  TabToOEM : TCvTable =(
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$AD,#$9B,#$9C,#$20,#$9D,#$B3,#$15,#$20,#$20,#$20,#$AE,#$AA,#$C4,#$20,#$20,
    #$F8,#$F1,#$FD,#$FC,#$20,#$E6,#$E3,#$20,#$20,#$20,#$20,#$AF,#$AC,#$AB,#$20,#$A8,
    #$41,#$41,#$41,#$41,#$8E,#$8F,#$92,#$80,#$45,#$90,#$45,#$45,#$49,#$49,#$49,#$49,
    #$44,#$A5,#$4F,#$4F,#$4F,#$4F,#$99,#$20,#$20,#$55,#$55,#$55,#$9A,#$59,#$20,#$E1,
    #$85,#$61,#$83,#$61,#$84,#$86,#$91,#$87,#$8A,#$82,#$88,#$89,#$8D,#$A1,#$8C,#$8B,
    #$20,#$A4,#$95,#$A2,#$93,#$6F,#$94,#$F6,#$20,#$97,#$A3,#$96,#$81,#$79,#$20,#$98);

// convert ANSI string to OEM - required for Zip
function StrToOEM (s : string) : string;
var
  i  : integer;
begin
  for i:=1 to length(s) do begin
    if s[i]>=#128 then s[i]:=TabToOEM[ord(s[i])-128];
    end;
  StrToOEM:=s;
  end;

{ ---------------------------------------------------------------- }
const
  TabToANSI : TCvTable =(
    #$C7,#$FC,#$E9,#$E2,#$E4,#$E0,#$E5,#$E7,#$EA,#$EB,#$E9,#$E8,#$EF,#$EE,#$C4,#$C5,
    #$C9,#$E6,#$C6,#$F4,#$F6,#$F2,#$FB,#$F9,#$FF,#$D6,#$DC,#$A2,#$A3,#$A5,#$20,#$20,
    #$E1,#$ED,#$F3,#$FA,#$F1,#$D1,#$61,#$6F,#$BF,#$20,#$AC,#$BD,#$BC,#$A1,#$AB,#$BB,
    #$20,#$20,#$20,#$A6,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$AD,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$DF,#$B6,#$20,#$20,#$20,#$B5,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,#$20,
    #$20,#$B1,#$20,#$20,#$20,#$20,#$F7,#$20,#$B0,#$20,#$B7,#$20,#$B3,#$B2,#$20,#$20);

    (* Umwandeln eines OEM-Strings nach ANSI *)
function StrToAnsi (const s : string) : string;
var
  i  : integer;
begin
  Result:=s;
  for i:=1 to length(Result) do begin
    if Result[i]>=#128 then Result[i]:=TabToANSI[ord(Result[i])-128]
    else if Result[i]=#$15 then Result[i]:=#$A7;
    end;
  end;

{ ------------------------------------------------------------------- }
function GetGzHeader (fSource   : TStream;
                      var fName : string;
                      var fTime : cardinal;
                      var fEnc  : boolean) : integer;
var
  Header : TGzHeader;
  n      : word;
  c      : char;
begin
  Result:=errOK;       // o.k.
  fName:=''; fTime:=0;
  try
    fSource.Read(Header,sizeof(TGzHeader));
    with Header do begin
      if Signature<>GzSignatur then begin
        Result:=errFileGZip; exit;
        end;
      fTime:=TimeStamp;
      with fSource do begin
        if Flags and gfExtraField <> 0 then begin   // skip extra field
          Read(n,2);              // length of extra field
          Position:=Position+n;   // adjust stream position
          end;
        if Flags and gfFileName <> 0 then begin   // read filename
          repeat
            Read(c,1);
            if c<>#0 then fName:=fName+c;
            until c=#0;
          end;
        if Flags and gfComment <> 0 then begin   // skip comment
          repeat
            Read(c,1);
            until c=#0;
          end;
        if Flags and gfHeaderCrc <> 0 then Read(n,2); // skip crc16
        fEnc:=Flags and gfEncrypted<>0;
        end;
      end;
  except
    Result:=errFileRead; exit; // error
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from stream *)
function GetGzInfo (sGz : TStream; var FileInfo : TGzFileInfo) : boolean;
begin
  with FileInfo do begin
    UnixTime:=DateTimeToUnixTime(Now);
    if GetGzHeader (sGz,Filename,UnixTime,Encrypted)=errOK then begin
      Attr:=FileGetAttr(Filename);
      with sGz do begin
        Seek (-4,soFromEnd);
        Read(USize,4);
        end;
      Result:=true;
      end
    else Result:=false;
    end;
  end;

{ ------------------------------------------------------------------- }
(* retrieve file info from gz-file *)
function GzFileInfo (const Filename: string; var FileInfo : TGzFileInfo) : boolean;
var
  sFile : TFileStream;
begin
  Result:=false;
  try
    sFile:= TFileStream.Create(Filename,fmOpenRead+fmShareDenyNone);
    Result:=GetGzInfo(sFile,FileInfo);
  finally
    sFile.Free;
    end;
  end;

function IsGzFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=GzExt;
  end;

function IsEncFile (const Filename : string) : boolean;
begin
  Result:=AnsiLowercase(ExtractFileExt(Filename))=CrExt;
  end;

function ZipCrc32(Crc : cardinal; const buf; len: Integer): cardinal;
begin
  Result:=Crc32 (Crc,pchar(@buf),len);
  end;

{ ------------------------------------------------------------------- }
constructor TCopyThread.Create (const ASourceName,ADestName : string;
                                ASuspend : Boolean; ABufSize : integer;
                                APriority : TThreadPriority);
begin
  inherited Create (ASuspend);
  SourceName:=ASourceName; 
  DestName:=ADestName;
  Priority:=APriority;
  FOnProgress:=nil;
  FCheckTime:=true;
  FCopyCrTm:=false;
  FCopyStorage:=false;
  FShellCopy:=[];
  FVerify:=false;
  FError:=errOK; FErrFlag:=errCopy;
  FBufSize:=ABufSize;
  SetLength(FBuffer,FBufSize);
  FUserBreak:=false;
  UTotal:=0; CTotal:=0;
  // init checksum
  FCrc:=Crc32(0,nil,0);
  FCompLevel:=zcDefault; FCompType:=zsDefault;
  FCrypting:=nil;
  end;

destructor TCopyThread.Destroy;
begin
  FBuffer:=nil;
  if assigned(FCrypting) then FCrypting.OnProgress:=nil;
  inherited Destroy;
  end;

function TCopyThread.GetDone : boolean;
begin
  Result:=Terminated;
  end;

procedure TCopyThread.SetProgress (ACallBack : TProgressEvent);
begin
  FOnProgress:=ACallBack;
  if assigned(FCrypting) then FCrypting.OnProgress:=ACallBack;
  end;

procedure TCopyThread.EndThread (Error : integer);
begin
  if Error<>errOK then FError:=FErrFlag or Error;
  Terminate;
  end;

procedure TCopyThread.CancelThread; 
begin
  FUserBreak:=true;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CopyStream (fSource,fDest : TStream; SLength : int64) : integer;
var
  NRead,NWrite,ec,fb : integer;
begin
  ec:=errOK;
  repeat
    if SLength<FBufSize then fb:=SLength else fb:=FBufSize;
    try
      NRead:=sSource.Read(FBuffer[0],fb);
      if NRead<fb then ec:=errError+errFileRead   // z.B. wenn "sSource" gelockt ist
      else begin
        UTotal:=UTotal+NRead;        // total number of bytes read
        DoProgress(acCopy,UTotal);
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        end;
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        NWrite:=sDest.Write(FBuffer[0],NRead);
        if NWrite<NRead then ec:=errError+errFileFull; // Ziel-Medium voll
      except
        ec:=errError+errFileWrite;
        end;
      if FUserBreak then ec:=errUserBreak;
      end;
    dec(Slength,NRead);
    until (SLength<=0) or (ec<>errOK);
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.GzCompress (fSource,fDest : TStream) : integer;
var
  sComp         : TZCompressionStream;
  NRead         : cardinal;
  ec            : integer;
  fp            : int64;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible compression
  sComp:=TZCompressionStream.Create(fDest,FCompLevel,defWindowBits,defMemLevel,FCompType);
  fp:=fDest.Position;
  repeat
    try
      NRead:=fSource.Read(FBuffer[0],FBufSize);
      UTotal:=UTotal+NRead;        // total number of bytes read
      DoProgress(acCompress,UTotal);
      FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        sComp.Write(FBuffer[0],NRead);
//   does not work with TZCompressionStream: NWrite is always equal to NRead
//        if NWrite<NRead then ec:=errError+errFileFull;  // dest disk full
      except
        ec:=errError+errFileWrite;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);
  if ec=errOK then sComp.Free else sComp.Break;
  if (ec=errOK) and (fSource.Size>0) and (fDest.Position<=fp) then
    ec:=errError+errFileFull;  // dest disk full
  Result:=ec;
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream) : integer;
var
  ALength : int64;
begin
  with fSource do ALength:=Size-Position-8; // gz-format
  Result:=GzUnCompress (fSource,fDest,ALength);
  end;

function TCopyThread.GzUnCompress (fSource,fDest : TStream; SLength : int64) : integer;
var
  sComp         : TZDecompressionStream;
  NRead,NWrite  : cardinal;
  RTotal        : int64;
  ec            : integer;
begin
  ec:=errOK;       // o.k.
  // setup for GZip compatible decompression
  sComp:=TZDecompressionStream.Create(fSource,SLength,defWindowBits);
  RTotal:=0;
  repeat
    try
      NRead:=sComp.Read(FBuffer[0],FBufSize);
    except
      ec:=errError+errFileRead;
      end;
    if ec=errOK then begin
      try
        FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        RTotal:=RTotal+NRead;
        DoProgress(acUnCompress,RTotal);
        if assigned(fDest) then NWrite:=fDest.Write(FBuffer[0],NRead)
        else NWrite:=NRead;
        UTotal:=UTotal+NWrite;        // total number of bytes written
        if (NWrite<NRead) then ec:=errError+errFileFull;    // dest disk full
      except
        ec:=errError+errFileWrite;
        end;
      if (ec=errOK) and FUserBreak then ec:=errUserBreak;
      end;
    until (NRead<FBufSize) or (ec<>errOK);;
  sComp.Free;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.DoProgress (AAction : TFileAction; ACount : int64);
begin
  if Assigned(FOnProgress) then begin
    FAction:=AAction; FCount:=ACount;
    Synchronize(UpdateProgress);
    end;
  end;

procedure TCopyThread.UpdateProgress;
begin
  FOnProgress(FAction,FCount);
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.OpenFiles (var FTime : TFileTimestamps; var Attr : word) : integer;
begin
  Result:=errOK;
  if length(SourceName)>MAX_PATH-1 then Result:=errError+errLongPath
  else if not ExistingFile (SourceName) then Result:=errError+errNotFound;
  if Result=errOK then begin
    if length(DestName)>MAX_PATH-1 then Result:=errError+errLongPath
    else if (length(DestName)=0) then Result:=errError+errFileCreate;
    end;
  if Result<>errOK then exit;
  FTime:=GetFileTimestamps(SourceName);
  Attr:=FileGetAttr(SourceName);
  try
    sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do begin
      Result:=errError+errFileOpen; Exit;
      end;
    end;
  // always overwrite destination, reset attributes
  if ExistingFile(DestName) then begin
    if FileSetAttr(DestName,faArchive)<>0 then begin
      Result:=errError+errFileWrite; Exit;
      end;
    end;
  try
    sDest:=TFileStream.Create(DestName,fmCreate);
  except
    on EFCreateError do Result:=errError+errFileCreate;
    end;
  end;

{ ------------------------------------------------------------------- }
procedure TCopyThread.ReleaseStreams;
begin
  if assigned(sSource) then try sSource.Free; except end;
  if assigned(sDest) then try sDest.Free; except end;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CloseFiles (FTime : TFileTimestamps; Attr : word) : integer;
begin
  Result:=errOK;
  FBytesRead:=sSource.Size;
  FBytesWritten:=sDest.Size;
  ReleaseStreams;
  if SetFileTimestamps(DestName,FTime,FCheckTime,FCopyCrTm)=0 then begin
    if FileSetAttr(DestName,Attr)>0 then Result:=errWarning+errFileAttr;
    end
  else Result:=errWarning+errFileTS;
  end;

{ ------------------------------------------------------------------- }
function TCopyThread.CopyFileStorage : integer;
var
  Summary : TFileSummary;
begin
  Result:=errOK;
  if FCopyStorage then begin
    if GetFileSummary(SourceName,Summary) then begin
      if not SetFileSummary(DestName,Summary) then Result:=errWarning+errStorage;
      end
    else Result:=errWarning+errStorage;
    end;
  end;

{ ------------------------------------------------------------------- }
// encrypt file
function TCopyThread.EncryptFile : integer;
var
  NRead    : cardinal;
  FTime    : TFileTimestamps;
  Attr     : word;
  ec       : integer;
begin
  ec:=OpenFiles (FTime,Attr);
  if (ec=errOK) then begin
    DoProgress(acEncrypt,-sSource.Size);
    with FCrypting do if WriteHeader (sDest) then begin
      repeat
        try
          NRead:=sSource.Read(FBuffer[0],FBufSize);
          UTotal:=UTotal+NRead;        // total number of bytes read
          DoProgress(acEncrypt,UTotal);
        except
          ec:=errError+errFileRead;
          end;
        if ec=errOK then begin
          try
            EncryptBlock (FBuffer[0],NRead);
            sDest.Write(FBuffer[0],NRead);
          except
            ec:=errError+errFileECrypt;
            end;
          if  FUserBreak then ec:=errUserBreak;
          end;
        until (NRead<FBufSize) or (ec<>errOK);
      if (ec=errOK) and not WriteTrailer (sDest) then ec:=errError+errFileECrypt;
      end
    else ec:=errError+errFileECrypt;
    if (ec=errOK) and (UTotal<>sSource.Size) then ec:=errError+errFileRead;  // z.B. wenn "sSource" gelockt ist
    if ec=errOK then ec:=CloseFiles (FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify encrypted file
function TCopyThread.VerifyEncFile : integer;
var
  ec        : integer;
begin
  ec:=errOK;
  try
    sDest:=TFileStream.Create(DestName,fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    DoProgress(acAddVerify,-UTotal);   // Reset progress indicator
    if not FCrypting.DecryptStream(sDest,nil,sDest.Size) then ec:=errError+errFileVerify;
    end;
  if assigned(sDest) then try sDest.Free; except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// decrypt file
function TCopyThread.DecryptFile : integer;
var
  FTime             : TFileTimestamps;
  Attr              : word;
  ec                : integer;
begin
  ec:=OpenFiles (FTime,Attr);
  if (ec=errOK) then begin
    if not FCrypting.DecryptStream(sSource,sDest,sSource.Size) then ec:=errError+errFileDCrypt;
    if ec=errOK then ec:=CloseFiles (FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// copy file
function TCopyThread.CopyFile : integer;
var
  FTime             : TFileTimestamps;
  ec                : integer;
  Attr              : word;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    DoProgress(acCopy,-sSource.Size);
    ec:=CopyStream(sSource,sDest,sSource.Size);
    if ec=errOk then ec:=CloseFiles (FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// copy using shell function "SHFileOperation" to include file infos
function TCopyThread.ShCopyFile : integer;
var
  FTime     : TFileTimestamps;
  ec        : integer;
  NRead     : cardinal;
  quiet     : boolean;
begin
  ec:=errOK;
  FTime:=GetFileTimestamps(SourceName);
  if length(SourceName)>MAX_PATH-1 then ec:=errError+errLongPath
  else if not ExistingFile (SourceName) then ec:=errError+errNotFound;
  if ec=errOK then begin
    if length(DestName)>MAX_PATH-1 then ec:=errError+errLongPath
    else if (length(DestName)=0) then ec:=errError+errFileCreate;
    end;
  if ec=errOK then begin
    UTotal:=LongFileSize(SourceName);
    quiet:=(scmQuiet in FShellCopy) or (UTotal<CMaxMemSize);
    DoProgress(acCopy,-1);   // Reset progress indicator
    if not ShellCopyFiles(Application.Handle,SourceName,DestName,'',quiet) then
       ec:=errError+errFileWrite;
    FBytesWritten:=UTotal;
    end;
  if FVerify and (ec=errOK) then begin   //Checksumme ermitteln
    UTotal:=0;
    try
      sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
    except
      on EFOpenError do ec:=errError+errFileVerify;
      end;
    if ec=errOK then begin
      DoProgress(acAddVerify,-sSource.Size);   // Reset progress indicator
      FCrc:=Crc32(0,nil,0);
      repeat
        try
          NRead:=sSource.Read(FBuffer[0],FBufSize);
          UTotal:=UTotal+NRead;
          DoProgress(acAddVerify,UTotal);
          FCrc:=ZipCrc32(FCrc,FBuffer[0],NRead);
        except
          ec:=errError+errFileVerify; NRead:=0;
          end;
        if FUserBreak then ec:=errUserBreak;
        until (NRead<FBufSize) or (ec<>errOK);
      sSource.Free;
      end;
    end;
  if FCopyCrTm and (SetFileTimestamps(DestName,FTime,FCheckTime,FCopyCrTm)<>0) then ec:=errWarning+errFileTS;
  Result:=ec;
  end;

// verify check sum of copied file
function TCopyThread.VerifyFile : integer;
var
  ec        : integer;
  NRead,Crc : cardinal;
  RTotal    : int64;
begin
  ec:=errOK; RTotal:=0;
  try
    sDest:=TFileStream.Create(DestName,fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    DoProgress(acAddVerify,-UTotal);   // Reset progress indicator
    Crc:=Crc32(0,nil,0);
    repeat
      try
        NRead:=sDest.Read(FBuffer[0],FBufSize);
        RTotal:=RTotal+NRead;
        DoProgress(acAddVerify,RTotal);
        Crc:=ZipCrc32(Crc,FBuffer[0],NRead);
      except
        ec:=errError+errFileVerify; NRead:=0;
        end;
      if FUserBreak then ec:=errUserBreak;
      until (NRead<FBufSize) or (ec<>errOK);
    if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
    end;
  if assigned(sDest) then try sDest.Free; except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TCopyThread.Execute;
begin
  if scmCopy in FShellCopy then EndThread (ShCopyFile)
  else EndThread (CopyFile);
  end;

{ TEncryptThread ---------------------------------------------------------------- }
constructor TEncryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errEnCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TEncryptThread.Execute;
begin
  if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    EndThread (EncryptFile);
    end
  else EndThread (CopyFile);
  end;

{ TDecryptThread ---------------------------------------------------------------- }
constructor TDecryptThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                        ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  if assigned(FCrypting) then FErrFlag:=errDeCrypt;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TDecryptThread.Execute;
begin
  if assigned(FCrypting) then EndThread (DecryptFile)
  else EndThread (CopyFile);
  end;

{ TGZipThread ------------------------------------------------------------------- }
constructor TGZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize, APriority);
  FCrypting:=AEncryption;
  FErrFlag:=errGZip;
  end;

{ ------------------------------------------------------------------- }
// write gz header
function TGZipThread.GzWriteHeader (fDest : TStream;
                                    fName : string;
                                    fMethode,
                                    fTime : cardinal;
                                    fEnc  : boolean) : integer;
var
  Header : TGzHeader;
  n      : integer;
begin
  Result:=errOK;       // o.k.
  n:=length(fname);
  with Header do begin
    Signature:=GzSignatur;
    Methode:=fMethode;
    if n>0 then Flags:=gfFileName else Flags:=0;
    if fEnc then Flags:=Flags or gfEncrypted;
    TimeStamp:=fTime;
    XFlags:=0;
    OpSys:=0;
    end;
  try
    if fDest.Write(Header,sizeof(TGzHeader))<sizeof(TGzHeader) then begin
      Result:=errError+errFileFull;  // dest disk full
      Exit;
      end;
    if n>0 then begin   // write filename
      if fDest.Write(fName[1],n+1)<n+1 then begin
        Result:=errError+errFileFull;  // dest disk full
        end;
      end;
  except
    Result:=errError+errFileWrite; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// write gz trailer
function TGZipThread.GzWriteTrailer (fDest : TStream; ACrc,ASize : cardinal) : integer;
begin
  Result:=errOK;       // o.k.
  try
    if fDest.Write(ACrc,4)<4 then Result:=errError+errFileFull  // dest disk full
    else if fDest.Write(ASize,4)<4 then Result:=errError+errFileFull;  // dest disk full
  except
    Result:=errError+errFileWrite; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// compress file
function TGZipThread.CompressFile : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  fp            : int64;
  Attr          : word;
  sTmp          : TStream;
  ec            : integer;
  FEnc,tf       : boolean;
  tzn           : string;
begin
  ec:=OpenFiles (FTime,Attr);
  if ec=errOK then begin
    FEnc:=assigned(FCrypting);
    // unix time for gzip
    utime:=FileTimeToUnixTime(FTime.LastWriteTime);
    ec:=GzWriteHeader(sDest,ExtractFilename(SourceName),Z_DEFLATED,utime,FEnc);
    fp:=sDest.Position;
    if FEnc then begin
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=TempDirectory+tmpEncName;
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      end
    else sTmp:=sDest;
    // compress file
    if ec=errOK then begin
      DoProgress(acCompress,-sSource.Size);   // Reset progress indicator
      ec:=GzCompress(sSource,sTmp);
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then ec:=errError+errFileRead;
        // encrypt file
        if FEnc then begin
          if (ec=errOK) then begin
            DoProgress(acEncrypt,-sTmp.Size);   // Reset progress indicator
            sTmp.Seek(0,soFromBeginning);
            if (ec=errOK) and not FCrypting.EncryptStream(sTmp,sDest) then
              ec:=errError+errFileECrypt;
            end;
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        if ec=errOK then begin
          CTotal:=sDest.Position-fp;
          if UTotal>MaxCardinal then UTotal:=MaxCardinal;
          if FEnc then ec:=GzWriteTrailer (sDest,0,TInt64(UTotal).Lo)
          else ec:=GzWriteTrailer (sDest,FCrc,TInt64(UTotal).Lo);
          end;
        end;
      end;
    if ec=errOK then ec:=CloseFiles (FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

// verify check sum of copied file
function TGZipThread.VerifyGzFile : integer;
var
  fn            : string;
  ft,Crc        : cardinal;
  FEnc          : boolean;
  n             : int64;
  ec            : integer;
begin
  ec:=errOK;
  try
    sDest:=TFileStream.Create(DestName,fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileVerify;
    end;
  if ec=errOK then begin
    DoProgress(acAddVerify,-CTotal);   // Reset progress indicator
    CrC:=FCrc;
    ec:=GetGzHeader (sDest,fn,ft,FEnc);
    FCrc:=Crc32(0,nil,0);
    if ec=errOK then begin
      if FEnc then begin // decrypt file
        with sDest do n:=Size-Position-8;
        if not FCrypting.DecryptStream(sDest,nil,n) then ec:=errError+errFileVerify;
        end
      else begin
        ec:=GzUncompress(sDest,nil);
        if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
        end;
      end;
    end;
  if assigned(sDest) then try sDest.Free; except end;
  Result:=ec;
  end;

{ ------------------------------------------------------------------- }
// execute thread
procedure TGZipThread.Execute;
begin
  EndThread (CompressFile);
  end;

{ TWinCopyThread ------------------------------------------------------------------- }
constructor TWinCopyThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                   ACompress,AVerify,ASuspend : Boolean; ABufSize : integer;
                                   APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,AEncryption,ASuspend,ABufSize,APriority);
  FCompress:=ACompress;
  FVerify:=AVerify;
  if assigned(FCrypting) then FErrFlag:=errEncrypt
  else if FCompress then FErrFlag:=errGZip
  else FErrFlag:=errCopy;
  end;

procedure TWinCopyThread.Execute;
var
  ec : integer;
begin
  if scmCopy in FShellCopy then begin
    ec:=ShCopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end
  else if FCompress then begin
    ec:=CompressFile;
    if FVerify and (ec=errOK) then ec:=VerifyGzFile;
    end
  else if assigned(FCrypting) then begin
//    DestName:=DestName+CrExt;
    ec:=EncryptFile;
    if FVerify and (ec=errOK) then ec:=VerifyEncFile;
    end
  else begin
    ec:=CopyFile;
    if FVerify and (ec=errOK) then ec:=VerifyFile;
    end;
  if FCopyStorage and (ec=errOK) then ec:=CopyFileStorage;
  EndThread (ec);
  end;

{ TGUnZipThread ------------------------------------------------------------------- }
constructor TGUnZipThread.Create (const ASourceName,ADestName : string;  AEncryption : TEncryption;
                                  ASuspend : Boolean; ABufSize : integer; APriority : TThreadPriority);
begin
  inherited Create (ASourceName,ADestName,ASuspend,ABufSize,APriority);
  FCrypting:=AEncryption;
  FErrFlag:=errGUnzip;
  end;

{ ------------------------------------------------------------------- }
// read gz header
function TGUnZipThread.GzReadHeader (fSource   : TStream;
                                     var fName : string;
                                     var fTime : cardinal;
                                     var fEnc  : boolean) : integer;
begin
  Result:=GetGzHeader (fSource,fName,fTime,fEnc);
  if Result<>errOk then Result:=errError+Result;
  end;

{ ------------------------------------------------------------------- }
// read gz trailer
function TGUnZipThread.GzReadTrailer (fSource : TStream) : integer;
var
  aCrc,FSize : cardinal;
begin
  Result:=errOK;       // o.k.
  try
    fSource.Read(aCrc,4);
    fSource.Read(FSize,4);
    if (FCrc<>aCrc) or (TInt64(UTotal).Lo<>FSize) then begin
      Result:=errError+errFileCheck; exit; // checksum error
      end;
  except
    Result:=errError+errFileRead; // error
    end;
  end;

{ ------------------------------------------------------------------- }
// uncompress file
function TGUnZipThread.UncompressFile : integer;
var
  FTime         : TFileTimestamps;
  utime         : cardinal;
  ec            : integer;
  FName         : string;
  fp,n          : int64;
  Attr          : word;
  sTmp          : TStream;
  FEnc,tf       : boolean;
  tzn           : string;
begin
  ec:=errOK;
  if length(SourceName)>MAX_PATH-1 then ec:=errError+errLongPath
  else if not ExistingFile (SourceName) then ec:=errError+errFileOpen;
  if (ec=errOK) and (length(DestName)=0) then ec:=errError+errFileCreate;
  if (ec=errOK) then begin
    FTime:=GetFileTimestamps(SourceName);
    Attr:=FileGetAttr(SourceName);
    end;
  if ec=errOK then begin
    try
      sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
    except
      ec:=errError+errFileOpen;
      end;
    end;
  if ec=errOK then begin
    ec:=GzReadHeader(sSource,FName,utime,FEnc);
    if (ec=errOK) then begin
//      if length(FName)>0 then DestName:=DestName+FName
//      else DestName:=DestName+ChangeFileExt(ExtractFilename(SourceName),'');
      // always overwrite destination, reset attributes
      if ExistingFile(DestName) then begin
        if FileSetAttr(DestName,faArchive)<>0 then ec:=errError+errFileAttr;
        end;
      end;
    if ec=errOK then begin
      try
        sDest:=TFileStream.Create(DestName,fmCreate);
      except
        ec:=errError+errFileCreate;
        end;
      end;
    if (ec=errOK) and FEnc and not assigned(FCrypting) then ec:=errError+errFileDCrypt;
    if ec=errOK then begin
      fp:=sSource.Position;
      if FEnc then begin // decrypt file
        DoProgress(acDecrypt,-sSource.Size);
        tf:=sSource.Size>CMaxMemSize;
        if tf then begin
          tzn:=TempDirectory+tmpEncName;
          sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
          end
        else sTmp:=TMemoryStream.Create;
        with sSource do n:=Size-Position-8;
        if FCrypting.DecryptStream(sSource,sTmp,n) then begin
          DoProgress(acUnCompress,-sTmp.Size);
          sTmp.Seek(0,soFromBeginning);
          ec:=GzUncompress(sTmp,sDest,sTmp.size);
          end
        else ec:=errError+errFileDCrypt;
        sTmp.Free;
        if tf then DeleteFile(tzn);
        // Ignore trailer
        end
      else begin
        DoProgress(acUnCompress,-sSource.Size);
        ec:=GzUncompress(sSource,sDest);
        if (ec=errOK) then begin
          CTotal:=sSource.Position-fp;
          ec:=GzReadTrailer (sSource);
          end;
        end;
      end;
    if ec=errOK then ec:=CloseFiles (FTime,Attr) else ReleaseStreams;
    end
  else ReleaseStreams;
  Result:=ec;
  end;

procedure TGUnZipThread.Execute;
begin
  EndThread (UncompressFile);
  end;

{ ------------------------------------------------------------------- }
constructor TPkHeader.Create (const Fn : string; Ec,Ts,Chk : cardinal;
                              AA,AF,AM : word; Ofs,CSz,USz : int64; AE2 : boolean);
begin
  inherited Create;
  FName:=Fn; ErrCode:=Ec; Timestamp:=Ts;  CRC:=Chk;
  Attr:=AA; GpFlag:=AF; CompMethode:=AM;
  Offset:=Ofs; CSize:=CSz; USize:=USz;
  Zip64:=(Offset>MaxCardinal) or (CSize>MaxCardinal) or (USize>MaxCardinal);
  IsDir:=CSz=0; IsAe2:=AE2;
  end;

{ ------------------------------------------------------------------- }
(* generate error message *)
constructor EZipError.Create (const ErrString : string;
                              ErrNR     : integer);
begin
  inherited Create (ErrString);
  ZipError:=ErrNr or ErrZip;
  end;

{ TZip  ----------------------------------------------------}
// Common Zip object
constructor TZip.Create (const ZipName,Password : string;
                         AKeyMode,ABufSize,AMode : integer);
begin
  inherited Create (ZipName,AMode);
  FileList:=TStringList.Create;
  FileList.Sorted:=false;
  if ABufSize>0 then FBufferSize:=ABufSize
  else FBufferSize:=defBufferSize;
  FErrorFlag:=false;
  ZName:=ZipName;
  PBase:='';
  PkExtraCrypt:=defPkExtraCrypt;
  FEncrypt:=false;
  InitEncryption(Password,AKeyMode);
  FPriority:=tpNormal;
  FOnProgress:=nil;
  end;

destructor TZip.Destroy;
begin
  if FEncrypt then FEncryption.Free; // release encryption
  FileList.Free;
  inherited Destroy;
  end;

procedure TZip.InitEncryption (Password : string; AKeyMode : integer);
begin
  if FEncrypt then FEncryption.Free;
  FEncrypt:=length(Password)>0;
  if FEncrypt then begin   // initialize encryption
    FEncryption:=TEncryption.Create(Password,AKeyMode,FBufferSize);
    PkExtraCrypt.Mode:=FEncryption.KeyMode;
    end
  else FEncryption:=nil;
  end;

procedure TZip.SetProgress (ACallBack : TProgressEvent);
begin
  FOnProgress:=ACallBack;
  if FEncrypt then FEncryption.OnProgress:=ACallBack;
  end;

{ ------------------------------------------------------------------- }
// Read Zip archive
constructor TZipRead.Create (const ZipName,Password : string;
                             AKeyMode,ABufSize : integer);
begin
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmOpenRead);
  except
    on EFCreateError do begin
      raise EZipError.Create (Format(rsErrOpening,[ZipName]),errError+errFileOpen);
      end;
    end;
  end;

destructor TZipRead.Destroy;
var
  i : integer;
begin
  with FileList do for i:=0 to Count-1 do Objects[i].Free;
  inherited Destroy;
  end;

// Read central directory
procedure TZipRead.ReadDirectory;
var
  pke   : TPkEndHeader;
  pkd   : TPkDirHeader;
  pkl   : TPkHeader;
  pkz   : TPKZip64CdRecord;
  pko   : TPKZip64CdLocator;
  s,sp       : string;
  usz,csz,fp : int64;
  i,ne,ec,
  el,k      : cardinal;
  pkelf     : TZip64ExtraField;
  Buffer    : array of byte;
begin
  Position:=Size-sizeof(pke);
  Read(pke,sizeof(pke));
  with pke do begin
    if Signatur<>PkEndSignatur then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
      Exit;
      end;
    if (pke.Offset=MaxCardinal) or (pke.DirSize=MaxCardinal) or (pke.TotalEntries=MaxWord) then begin // Zip64 extension
      Position:=Size-sizeof(pke)-sizeof(pko);  // read Zip64 end of central directory locator
      Read(pko,sizeof(pko));
      if pko.Signatur<>Pk64LocSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        exit;
        end;
      Position:=pko.Offset;                    // read Zip64 end of central directory record
      Read(pkz,sizeof(pkz));
      if pkz.Signatur<>Pk64RecSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Exit;
        end;
      with pkz do begin
        fp:=Offset;
        ne:=TotalEntries;
        end;
      end
    else with pke do begin
      fp:=Offset;
      ne:=TotalEntries;
      end;
    end;
  Position:=fp;
  sp:='';
  for i:=0 to ne-1 do begin  // read directory entries
    Read(pkd,sizeof(pkd));
    ec:=errOK;
    with pkd do begin
      if Signatur<>PkDirSignatur then begin
        raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errSignature);
        Exit;
        end;
      if not ((VersExtr=VersDef) or (VersExtr=VersZip64)) then ec:=errFormat;
      if not ((Method=zipNoComp) or(Method=zipDeflate) or (Method=zipEncrypted)) then ec:=errFormat;
      fp:=Offset;
      usz:=USize;
      csz:=CSize;
      SetLength(s,FNLength);
      Read(s[1],FNLength);
      if ExtraLength>0 then begin      // read extra fields
        el:=ExtraLength;
        while el>0 do begin  // read extra fields
          Read(pkelf,SizeOf(TZip64ExtraField));
          with pkelf do begin
            SetLength(Buffer,ESize);
            Read(Buffer[0],ESize);
            k:=0;
            if Tag=ExtraZip64 then begin
              if USz=MaxCardinal then begin
                Move(Buffer[k],USz,8);
                inc(k,8);
                end;
              if (CSz=MaxCardinal) and (ESize>k) then begin
                Move(Buffer[k],CSz,8);
                inc(k,8);
                end;
              if (fp=MaxCardinal) and (ESize>k) then begin
                Move(Buffer[k],fp,8);
                end;
              end;
            el:=el-SizeOf(TZip64ExtraField)-ESize;
            end;
          end;
        end;
      pkl:=TPkHeader.Create(s,ec,FTimeStamp,CRC,ExtAttr,Flag,Method,fp,CSz,USz,true);
      FileList.AddObject(UnixPathToDosPath(StrToAnsi(s)),pkl)
      end;
    end;
  end;

function TZipRead.Filename(Index : integer) :string;
begin
  with FileList do if (Index>=0) and (Index<Count) then Result:=Strings[Index]
  else Result:='';
  end;

function TZipRead.IsEncrypted(Index : integer) : boolean;
begin
  Result:=false;
  with FileList do if (Index>=0) and (Index<Count) then
    Result:=((Objects[Index] as TPkHeader).GpFlag and zipEncryptFlag)<>0;
  end;

procedure TZipRead.ExtractFromZip (Offset : int64; TimeStamp,Attr : integer; const DestFilename : string);
var
  ec   : integer;
begin
  if length(DestFilename)>MAX_PATH-1 then ec:=errLongPath
  else begin
    Position:=Offset;
    ZipThread:=TZipReadThread.Create(DestFilename,FEncryption,self,false,FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        until Done;
      ec:=ErrorCode and errAllCodes;
      Free;
      end;
    ZipThread:=nil;
    end;
  if ec=errOK then begin
    if SetFileAge(DestFileName,TimeStamp)=0 then begin
      if FileSetAttr(DestFileName,Attr)>0 then
        raise EZipError.Create (Format(rsErrSetAttr,[DestFileName]),errWarning+errFileAttr);
      end
    else raise EZipError.Create (Format(rsErrTimeStamp,[DestFileName]),errWarning+errFileTS);
    end
  else begin
  // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

procedure TZipRead.ExtractFileFromZip (Index : integer; const DestFilename : string);
begin
  if assigned(FileList) then if (Index>=0) and (Index<FileList.Count) then
    with (FileList.Objects[Index] as TPkHeader) do begin
    if ((GpFlag and zipEncryptFlag)<>0) and not assigned(FEncryption) then begin
      raise EZipError.Create (Format(rsErrReading,[ZName]),errError+errFileDCrypt);
      Exit;
      end;
    ExtractFromZip(Offset,TimeStamp,Attr,DestFileName);
    end;
  end;

procedure TZipRead.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

{ ------------------------------------------------------------------- }
// Create ZIP archive
constructor TZipWrite.Create (const ZipName,BasicDirectory,Password : string;
                              AKeyMode,ABufSize : integer);
begin
  try
    inherited Create (ZipName,Password,AKeyMode,ABufSize,fmCreate);
    PBase:=IncludeTrailingPathDelimiter(BasicDirectory);
  except
    on EFCreateError do begin
      raise EZipError.Create(Format(rsErrCreating,[ZipName]),errError+errFileCreate);
      end;
    end;
  FBytesRead:=0;
  FOnVerifyProgress:=nil;
  end;

//  Write directory and final block of ZIP archive
destructor TZipWrite.Destroy;
var
  i : integer;
begin
  with FileList do for i:=0 to Count-1 do Objects[i].Free;
  inherited Destroy;
  end;

procedure TZipWrite.WriteDirectory;
var
  pke : TPkEndHeader;
  pkd : TPkDirHeader;
  pkl : TPkHeader;
  pkz : TPKZip64CdRecord;
  pko : TPKZip64CdLocator;
  fp  : int64;
  ez64l,
  i,n : integer;
  lef : TPKExtraLocalField;
begin
  fp:=Size;
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    pkl:=Objects[i] as TPkHeader;
    with pkd do begin
      Signatur:=PkDirSignatur;
      Flag:=pkl.GpFlag;
      Method:=pkl.CompMethode;
      FTimeStamp:=pkl.TimeStamp;
      FNLength:=length(pkl.FName);
      if pkl.Zip64 then begin
        VersMade:=VersZip64+VersDos;
        VersExtr:=VersZip64+VersDos;
        with lef do begin
          Tag:=ExtraZip64;
          n:=0;
          if pkl.USize>MaxCardinal then begin
            USize:=MaxCardinal;
            Buffer[n]:=pkl.USize;
            inc(n);
            end
          else USize:=TInt64(pkl.USize).Lo;
          if pkl.CSize>MaxCardinal then begin
            CSize:=MaxCardinal;
            Buffer[n]:=pkl.CSize;
            inc(n);
            end
          else CSize:=TInt64(pkl.CSize).Lo;
          if pkl.Offset>MaxCardinal then begin
            Offset:=MaxCardinal;
            Buffer[n]:=pkl.Offset;
            inc(n);
            end
          else Offset:=TInt64(pkl.Offset).Lo;
          ESize:=8*n;
          ez64l:=4+ESize;
          ExtraLength:=ez64l;
          end;
        end
      else begin
        VersMade:=VersDef+VersDos;
        VersExtr:=VersDef;
        CSize:=TInt64(pkl.CSize).Lo;
        USize:=TInt64(pkl.USize).Lo;
        Offset:=TInt64(pkl.Offset).Lo;
        ExtraLength:=0;
        end;
      if FEncrypt then begin
        ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
        if pkl.IsAe2 then begin
          CRC:=0;
          PkExtraCrypt.Version:=VendorVersAE2;
          end
        else begin
          CRC:=pkl.CRC;
          PkExtraCrypt.Version:=VendorVersAE1
          end;
        end
      else CRC:=pkl.CRC;
      CommLength:=0;
      DiskStart:=0;
      IntAttr:=0;
      ExtAttr:=pkl.Attr;
      end;
    try
      Write(pkd,sizeof(pkd));
      Write(pkl.FName[1],pkd.FNLength);
      if pkl.Zip64 then Write(lef,ez64l);  // extra fields
      if FEncrypt then Write(PkExtraCrypt,sizeof(TPKExtraEncField));
    except
      raise EZipError.Create (Format(rsErrDirEntry,[ZName]),errError+errFileWrite);
      exit;
      end;
    end;
  with pke do begin
    Signatur:=PkEndSignatur;
    ThisDisk:=0;
    StartDisk:=0;
    if FileList.Count>MaxWord then begin
      ThisEntries:=MaxWord;
      TotalEntries:=MaxWord;
      end
    else begin
      ThisEntries:=FileList.Count;
      TotalEntries:=FileList.Count;
      end;
    CommLength:=0;
    end;  // check if file size > 4GB
  if (Position>MaxCardinal) or (FileList.Count>MaxWord) then begin
    // Zip64 end of central directory record
    with pkz do begin
      Signatur:=Pk64RecSignatur;
      Size:=sizeof(TPKZip64CdRecord)-12;
      VersMade:=VersZip64+VersDos;
      VersExtr:=VersZip64+VersDos;
      ThisDisk:=0;
      StartDisk:=0;
      ThisEntries:=FileList.Count;
      TotalEntries:=FileList.Count;
      DirSize:=Position-fp;
      Offset:=fp;
      end;
    with pke do begin
      if pkz.DirSize>MaxCardinal then DirSize:=MaxCardinal
      else DirSize:=pkz.DirSize;
      Offset:=MaxCardinal;
      end;
    fp:=Position;
    // Zip64 end of central directory locator
    with pko do begin
      Signatur:=Pk64LocSignatur;
      StartDisk:=0;
      Offset:=fp;
      TotalDisk:=1;
      end;
    try
      Write(pkz,sizeof(TPKZip64CdRecord));
      Write(pko,sizeof(TPKZip64CdLocator));
    except
      raise EZipError.Create (Format(rsErrDirEntry,[ZName]),errError+errFileWrite);
      exit;
      end;
    end
  else with pke do begin
    DirSize:=Position-fp;
    Offset:=TInt64(fp).Lo;
    end;
  try
    Write(pke,sizeof(pke));
  except
    raise EZipError.Create (Format(rsErrEndRecord,[ZName]),errError+errFileWrite);
    end;
  FBytesWritten:=Size;
  end;

procedure TZipWrite.Verify;  // Verify zip archive
var
  i   : integer;
  ok  : boolean;
begin
  ok:=true;
  if assigned(FOnVerifyProgress) then FOnVerifyProgress(acVerify,-FileList.Count);
  with FileList do for i:=0 to Count-1 do begin  // central directory structure
    if assigned(FOnVerifyProgress) then FOnVerifyProgress(acVerify,i);
    ok:=ok and VerifyFile(Objects[i] as TPkHeader);
    if not ok then Break;
    end;
  if not ok then begin
    raise EZipError.Create (Format(rsErrVerify,[ZName]),errError+errFileVerify);
    end;
  end;

//  Add directory name to archive
procedure TZipWrite.AddDir (DirName : string);
var
  ZfName,FName : string;
  ec         : integer;
  LFHeader   : TPKLocalHeader;
  fp         : int64;
begin
  if DirectoryExists (DirName) then begin
    FName:=ExtractRelativePath(PBase,DirName);
    ZfName:=StrToOEM(FName); // ZIP needs OEM characters (like DOS)
    ec:=errOK;
    with LFHeader do begin
      Signatur:=PKLocalSignatur;
      Vers:=VersDef;
      Flag:=0;
      Method:=zipNoComp;
      FTimeStamp:=DirectoryAge(ExcludeTrailingPathDelimiter(DirName));
      CRC:=0;
      FNLength:=length(ZfName);
      CSize:=0;
      USize:=0;
      ExtraLength:=0;
      end;
    fp:=Position;
    try
      Write(LFHeader,sizeof(TPKLocalHeader));         // local header
      Write(ZfName[1],LFHeader.FNLength);             // filename
      FileList.Addobject(FName,TPkHeader.Create(ZfName,0,LfHeader.FTimeStamp,0,
               FileGetAttr(DirName),LfHeader.Flag,LfHeader.Method,fp,0,0,true));
    except
      on EWriteError do ec:=errFileWrite;
      end;
    if ec<>errOK then begin
      // set error flag on write errors
      FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
      raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
      end;
    end;
  end;

procedure TZipWrite.RemLastDir;
begin
  with FileList do if Count>0 then begin
    with (Objects[Count-1] as TPkHeader) do if IsDir then begin
      Size:=Offset;
      Free;
      Delete(Count-1);
      end;
    end;
  end;

// Add file to archive
procedure TZipWrite.AddFile (SourceName : string);
var
  ZfName,FName : string;
  ec        : integer;
  utime     : cardinal;
  LFHeader  : TPKLocalHeader;
  lef       : TPKExtraLocalField;
  FileData  : TWin32FindData;
  ut,ct,fp  : int64;
  z64,zae2  : boolean;
  rf,fc,el  : cardinal;
begin
  if length(SourceName)>MAX_PATH-1 then ec:=errLongPath
  else if ExistingFile (SourceName) then begin
    FName:=ExtractRelativePath(PBase,SourceName);
    ZfName:=StrToOEM(FName); // ZIP needs OEM characters (like DOS)
    if GetFileData(SourceName,FileData) then begin
      utime:=GetFileAge(FileData);
      ut:=GetFileSize(FileData);
      FBytesRead:=FBytesRead+ut;
      z64:=ut>MaxSize;
      zae2:=ut<AE1Size;
      with LFHeader do begin
        Signatur:=PKLocalSignatur;
        if z64 then Vers:=VersZip64 else Vers:=VersDef;
        if FEncrypt then begin
          Flag:=zipEncryptFlag;
          Method:=zipEncrypted;
          end
        else begin
          Flag:=zipMaxDeflFlag;
          Method:=zipDeflate;
          end;
        FTimeStamp:=utime;
        CRC:=0;
        FNLength:=length(ZfName);
        if z64 then begin
          with lef do begin
            Tag:=ExtraZip64;
            ESize:=16;
            Buffer[0]:=0;
            Buffer[1]:=0;
            end;
          CSize:=MaxCardinal;
          USize:=MaxCardinal;
          el:=20;
          ExtraLength:=20;
          end
        else begin
          CSize:=0;
          USize:=0;
          ExtraLength:=0;
          end;
        if FEncrypt then begin
          ExtraLength:=ExtraLength+sizeof(TPKExtraEncField);
          with PkExtraCrypt do if zae2 then Version:=VendorVersAE2 else Version:=VendorVersAE1;
          end;
        end;
      fp:=Position;
      try
        Write(LFHeader,sizeof(TPKLocalHeader));         // local header
        Write(ZfName[1],LFHeader.FNLength);             // filename
        if z64 then Write(lef,el);                      // extra Zip64 field
        if FEncrypt then Write(PkExtraCrypt,sizeof(TPKExtraEncField));  // extra enc field
        ZipThread:=TZipWriteThread.Create(SourceName,FEncryption,self,false,FBufferSize,FPriority,FOnProgress);
        with ZipThread do begin
          repeat
            Application.ProcessMessages;
            Sleep(10);
            until Done;
          ec:=ErrorCode and errAllCodes;
          ut:=UTotal;
          ct:=CTotal;
          fc:=FCrc;
          Free;
          end;
        ZipThread:=nil;
        if ec=errOK then begin
          if FEncrypt and zae2 then Position:=fp+18    // position of compressed size
          else begin                // Crc if no encryption or AE-1
            Position:=fp+14;                // position of crc
            Write(fc,sizeof(cardinal));
            end;
          if z64 then begin
            // write 8 byte values to extra field
            Position:=fp+sizeof(TPKLocalHeader)+LFHeader.FNLength+4;
            Write (ut,sizeof(int64));
            Write (ct,sizeof(int64));
            end
          else begin
            // write 4 byte values
            rf:=TInt64(ct).Lo;
            Write(rf,sizeof(cardinal));
            rf:=TInt64(ut).Lo;
            Write(rf,sizeof(cardinal));
            end;
          Seek (0,soEnd);
          FileList.Addobject(FName,TPkHeader.Create(ZfName,ec,utime,fc,
                   FileGetAttr(SourceName),LfHeader.Flag,LfHeader.Method,fp,ct,ut,zae2));
          end
      except
        on EWriteError do ec:=errFileWrite;
        end;
      if ec<>errOK then Size:=fp;    // reset on error
      end
    else ec:=errFileRead;
    end
  else ec:=errNotFound;
  if ec<>errOK then begin
    // set error flag on write errors
    FErrorFlag:=(ec=errFileWrite) or (ec=errFileFull);
    raise EZipError.Create (GetCopyErrMsg(errError+ec),errError+ec);
    end;
  end;

{ TPkHeader = class(TObject)
    FName           : string;  // Original filename (OEM)
    TimeStamp,CRC   : cardinal;
    GpFlag,Attr,
    CompMethode     : word;
    Offset,
    CSize,USize     : int64;
    IsDir,
    Zip64           : boolean;
}
function TZipWrite.VerifyFile (PkHeader : TPkHeader) : boolean;
var
  LFHeader  : TPKLocalHeader;
  Buffer    : array of byte;
begin
  with PkHeader do if CSize>0 then begin
    Position:=Offset;
    // Header berlesen
    Read(LFHeader,sizeof(TPKLocalHeader));         // local header
    SetLength(Buffer,LFHeader.FNLength);
    Read(Buffer[0],LFHeader.FNLength);             // filename
    if LFHeader.ExtraLength>0 then begin
      SetLength(Buffer,LFHeader.ExtraLength);
      Read(Buffer[0],LFHeader.ExtraLength);             // filename
      end;
    ZipThread:=TZipVerifyThread.Create(FEncryption,Crc,self,CSize,false,FBufferSize,FPriority,FOnProgress);
    with ZipThread do begin
      repeat
        Application.ProcessMessages;
        Sleep(10);
        until Done;
      Result:=ErrorCode=errOK;
      Free;
      end;
    end
  else Result:=true;
  end;

procedure TZipWrite.Cancel;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then CancelThread;
  end;

procedure TZipWrite.Suspend;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Suspend;
  end;

procedure TZipWrite.Resume;
begin
  if assigned(ZipThread) then with ZipThread do if not Done then Resume;
  end;

{ TZipWriteThread ------------------------------------------------------------------- }
constructor TZipWriteThread.Create (const AFileName : string; AEncryption : TEncryption;
                               ZStream : TStream; ASuspend : Boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create (AFileName,'',AEncryption,ASuspend,ABufSize,APriority);
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

{ ------------------------------------------------------------------- }
// add file to zip file
function TZipWriteThread.AddToZip : integer;
var
  fp     : int64;
  sTmp   : TStream;
  ec     : integer;
  tf     : boolean;
  tzn    : string;
begin
  ec:=errOK;
  fp:=sDest.Position;
  // compress file
  try
    sSource:=TFileStream.Create(SourceName,fmOpenRead+fmShareDenyNone);
  except
    on EFOpenError do ec:=errError+errFileOpen;
    end;
  if ec=errOK then begin
    DoProgress(acCompress,-sSource.Size);   // Reset progress indicator
    if assigned(FCrypting) then begin
      tf:=sSource.Size>UMaxMemSize;
      if tf then begin
        tzn:=TempDirectory+tmpEncName;
        sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
        end
      else sTmp:=TMemoryStream.Create;
      ec:=GzCompress(sSource,sTmp);
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then ec:=errError+errFileRead;
        end;
      DoProgress(acEncrypt,-sTmp.Size);   // Reset progress indicator
      // encrypt file
      if (ec=errOK) then begin
        sTmp.Seek(0,soFromBeginning);
        if not FCrypting.EncryptStream(sTmp,sDest) then ec:=errError+errFileECrypt;
        end;
      sTmp.Free;
      if tf then DeleteFile(tzn);
      end
    else begin
      ec:=GzCompress(sSource,sDest);
      if ec=errOK then begin
        // z.B. wenn "sSource" gelockt ist (siehe LockFile)
        if UTotal<>sSource.Size then ec:=errError+errFileRead;
        end;
      end;
    try sSource.Free; except end;
    CTotal:=sDest.Position-fp;
    end
  else if assigned(sSource) then try sSource.Free; except end;
  Result:=ec;
  end;

procedure TZipWriteThread.Execute;
begin
  EndThread(AddToZip);
  end;

{ ------------------------------------------------------------------- }
constructor TPkFileRef.Create (Ofs : int64);
begin
  inherited Create;
  Offset:=Ofs;
  end;

{ TZipVerifyThread ------------------------------------------------------------------- }
constructor TZipVerifyThread.Create (AEncryption : TEncryption; ACrc : cardinal;
                               ZStream : TStream; ALength : int64; ASuspend : Boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create ('','',AEncryption,ASuspend,ABufSize,APriority);
  OnProgress:=AOnProgress;
  sDest:=TFileStream(ZStream);
  FLength:=ALength;
  FCrc:=ACrc;
  FErrFlag:=errZip;
  end;

function TZipVerifyThread.VerifyZippedFile : integer;
var
  ec  : integer;
  crc : cardinal;
begin
  Crc:=FCrc;
  FCrc:=Crc32(0,nil,0);
  if assigned(FCrypting) then begin // decrypt file
    if not FCrypting.DecryptStream(sDest,nil,FLength) then ec:=errError+errFileVerify
    else ec:=errOK;
    end
  else begin
    ec:=GzUncompress(sDest,nil,FLength);
    if (ec=errOK) and (FCrc<>Crc) then ec:=errError+errFileVerify;
    end;
  Result:=ec;
  end;

procedure TZipVerifyThread.Execute;
begin
  EndThread(VerifyZippedFile);
  end;

{ TZipReadThread ------------------------------------------------------------------- }
constructor TZipReadThread.Create (const AFileName : string; AEncryption : TEncryption;
                               ZStream : TStream; ASuspend : Boolean;
                               ABufSize : integer; APriority : TThreadPriority;
                               AOnProgress : TProgressEvent);
begin
  inherited Create ('',AFileName,AEncryption,ASuspend,ABufSize,APriority);
  OnProgress:=AOnProgress;
  sSource:=TFileStream(ZStream);
  FErrFlag:=errZip;
  end;

function TZipReadThread.ReadZipLocalHeader (var ZipInfo : TZipFileInfo) : boolean;
var
  LFHeader  : TPKLocalHeader;
  el,k      : cardinal;
  pkelf     : TZip64ExtraField;
  Buffer    : array of byte;
begin
  sSource.Read(LFHeader,sizeof(TPKLocalHeader));         // local header
  with LFHeader do if Signatur=PKLocalSignatur then begin
    SetLength(ZipInfo.Filename,FNLength);
    sSource.Read(ZipInfo.Filename[1],FNLength);
    ZipInfo.TimeStamp:=FTimeStamp;
    ZipInfo.Crc:=Crc;
    ZipInfo.CSize:=CSize;
    ZipInfo.USize:=USize;
    ZipInfo.Flag:=Flag;
    ZipInfo.Method:=Method;
    if ExtraLength>0 then begin      // read extra fields
      el:=ExtraLength;
      while el>0 do begin  // read extra fields
        sSource.Read(pkelf,SizeOf(TZip64ExtraField));
        with pkelf do begin
          SetLength(Buffer,ESize);
          sSource.Read(Buffer[0],ESize);
          k:=0;
          if Tag=ExtraZip64 then begin
            if USize=MaxCardinal then begin
              Move(Buffer[k],ZipInfo.USize,8);
              inc(k,8);
              end;
            if (CSize=MaxCardinal) and (ESize>k) then begin
              Move(Buffer[k],ZipInfo.CSize,8);
              end;
            end;
          el:=el-SizeOf(TZip64ExtraField)-ESize;
          end;
        end;
      Buffer:=nil;
      end;
    Result:=true;
    end
  else Result:=false;
  end;

function TZipReadThread.ReadFromZip : integer;
var
  ZipInfo       : TZipFileInfo;
  ec            : integer;
  sTmp          : TStream;
  tf            : boolean;
  tzn           : string;
begin
  ec:=errOK;
  if length(DestName)>MAX_PATH-1 then ec:=errError+errLongPath
  else if (length(DestName)=0) then ec:=errError+errFileCreate;
  if ec=errOK then begin
    try
      sDest:=TFileStream.Create(DestName,fmCreate);
    except
      ec:=errError+errFileCreate;
      end;
    end;
  if (ec=errOK) then begin
    if ReadZipLocalHeader(ZipInfo) then begin
      if (ZipInfo.Flag and zipEncryptFlag)<>0 then begin  // decrypt
        if not assigned(FCrypting) then ec:=errError+errFileDCrypt
        else begin
          DoProgress(acDecrypt,-ZipInfo.CSize);
          tf:=ZipInfo.CSize>CMaxMemSize;
          if tf then begin
            tzn:=TempDirectory+tmpEncName;
            sTmp:=TFileStream.Create(tzn,fmCreate); // write to disk
            end
          else sTmp:=TMemoryStream.Create;
          if FCrypting.DecryptStream(sSource,sTmp,ZipInfo.CSize) then begin
            DoProgress(acUnCompress,-sTmp.Size);
            sTmp.Seek(0,soFromBeginning);
            ec:=GzUncompress(sTmp,sDest,sTmp.size);
            end
          else ec:=errError+errFileDCrypt;
          sTmp.Free;
          if tf then DeleteFile(tzn);
          end;
        end
      else begin
        DoProgress(acUnCompress,-ZipInfo.CSize);
        if ZipInfo.Method=zipDeflate then ec:=GzUncompress(sSource,sDest,ZipInfo.CSize)
        else ec:=CopyStream(sSource,sDest,ZipInfo.USize);
        if (ec=errOK) and (FCrc<>ZipInfo.Crc) then ec:=errError+errFileRead;
        end;
      end
    else ec:=errError+errFileRead;
    try sDest.Free; except end;
    end;
  Result:=ec;
  end;

procedure TZipReadThread.Execute;
begin
  EndThread(ReadFromZip);
  end;

end.
