(* Delphi Unit
   several procedures and functions for file and directory processing

    2010 J. Rathlev, IEAP, Uni-Kiel, (rathlev(a)physik.uni-kiel.de)

   The contents of this file may be used under the terms of the
   Mozilla Public License ("MPL") or
   GNU Lesser General Public License Version 2 or later (the "LGPL")

   Software distributed under this License is distributed on an "AS IS" basis,
   WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License for
   the specific language governing rights and limitations under the License.

   Vers. 1 - Feb. 2010
   *)

unit FileUtils;

interface

uses Windows, Classes, SysUtils;

const
  FILE_WRITE_ATTRIBUTES = $0100;

type
  TInt64 = record
    case integer of
    0: (AsInt64 : int64);
    1: (Lo, Hi  : Cardinal);
    2: (Cardinals: array [0..1] of Cardinal);
    3: (Words: array [0..3] of Word);
    4: (Bytes: array [0..7] of Byte);
    5 :(FileTime : TFileTime);
    end;

  TFileTimestamps = record
    Valid          : boolean;
    CreationTime,
    LastAccessTime,
    LastWriteTime  : TFileTime;
    end;

  TFileInfo = record
    Name     : string;
    DateTime : TDateTime;
    Size     : int64;
    Attr     : integer;
    end;

{ ---------------------------------------------------------------- }
  ECopyError = class(EInOutError)
  public
    CopyError : integer;
    constructor Create (const ErrString : string; ErrNR : integer);
    end;

// search abs. Position in stream, even if >= 2^32
procedure SeekPosition (sFile : TFileStream; Pos : int64; Origin : word);

{ ---------------------------------------------------------------- }
// get time (UTC) of last file write
function GetFileLastWriteTime(const FileName: string): TFileTime;
function GetFileLastWriteDateTime(const FileName: string): TDateTime;


// set time (UTC) of last file write
function SetFileLastWriteTime(const FileName: string; FileTime : TFileTime;
                              CheckTime : boolean) : integer;

// get file timestamps (UTC)
function GetFileTimestamps(const FileName: string): TFileTimestamps;

// get directory timestamps (UTC)
function GetDirTimestamps(const DirName: string): TFileTimestamps;

// get file or directory timestamps (UTC) from FindData
function GetTimestampsFromFindData(const FindData : TWin32FindData) : TFileTimestamps;

// set file timestamps (UTC)
function SetFileTimestamps (const FileName: string; Timestamps : TFileTimestamps;
                            CheckTime,SetCreationTime : boolean) : integer;

// convert Filetime to Delphi time (TDateTime)
function FileTimeToDateTime (ft : TFileTime) : TDateTime;

// convert Delphi time (TDateTime) to Filetime
function DateTimeToFileTime (dt : TDateTime) : TFileTime;

// FileDateToDateTime with error check
function DateTimeFromFileDate(FileDate : integer) : TDateTime;

// convert Delphi time to Unix time and reverse
function DateTimeToUnixTime (dt : TDateTime) : cardinal;
function UnixTimeTodateTime (ut : cardinal) : TDateTime;

// convert Fileage to Unix time
function FileAgeToUnixTime (Age : cardinal) : cardinal;

// convert Filetime to Unix time and reverse
function FileTimeToUnixTime (ft : TFileTime) : cardinal;
function UnixTimeToFileTime (ut : cardinal) : TFileTime;

// convert Filetime to seconds
function FileTimeToSeconds (ft : TFileTime) : cardinal;

{ ---------------------------------------------------------------- }
// convert file attribute to string
function FileAttrToString(Attr : word) : string;

// read file infos (WIN32_FIND_DATA)
function GetFileData(FileName : string; var FileData : TWin32FindData) : boolean;


{ ---------------------------------------------------------------- }
// get file size as int64
function GetFileSize (const FileData: TWin32FindData) : int64;
function LongFileSize (const FileName : string) : int64;

// Convert SearchRec to FileInfo
function SearchRecToFileInfo (SearchRec : TSearchRec) : TFileInfo;

{ ---------------------------------------------------------------- }
// check if file exists
function ExistingFile (const FileName: string): boolean;

// Delete file even if readonly
function EraseFile(const FileName: string): Boolean;

// get DOS file age from file data
function GetFileAge (const FileData: TWin32FindData) : integer;

// Get time stamp of directory (see FileAge)
function DirectoryAge(const FileName: string): Integer;

// Get time stamp of file or directory (see FileSetDate)
function SetFileAge(const FileName: string; Age: Integer): Integer;
function SetDirectoryAge(const FileName: string; Age: Integer): Integer;

// Delete all matching files in directory
procedure DeleteMatchingFiles(const APath,AMask : string);

{ ---------------------------------------------------------------- }
// Prfe, ob Textdatei vorhanden
function ExistsFile (var f : TextFile) : boolean;

// Prfen, ob in einem Verzeichnis eine Datei existiert, die der Maske entspricht
function FileMatchesMask (Dir,Mask : string) : string;

{ ---------------------------------------------------------------- }
// copy file
procedure FileCopy (srcfilename,destfilename : String); overload;
procedure FileCopy (srcfilename,destfilename : String; Attr : word); overload;

// Copy file with timestamp *)
procedure CopyFileTS (const srcfilename, destfilename: String; BlockSize : integer);

{ ---------------------------------------------------------------- }
// kein spezielles Verzeichnis (selbst und eins zurck oder Verzeichnis, das mit StartChar beginnt)
function NotSpecialDir (const Name : string; StartChars : array of char) : boolean; overload;
function NotSpecialDir (const Name : string) : boolean; overload;

// Check if reparse point is recursive
function IsRecursiveParsePoint (const Path : string; Attr : integer; AllowJunctions : boolean) : boolean; overload;
function IsRecursiveParsePoint (const Path : string; AllowJunctions : boolean) : boolean; overload;

// Check if directory is empty
function IsEmptyDir (const Directory : string) : boolean;

// Dateien in einem Verzeichnis zhlen und Gesamtgre bestimmen
procedure CountFiles (const Base,Dir : string; IncludeSubDir : boolean;
                      var FileCount : cardinal; var FileSize : int64);
function DirFiles (const Directory : string; IncludeSubDir : boolean) : cardinal;
function DirSize (const Directory : string; IncludeSubDir : boolean) : int64;

(* Lsche ein Verzeichnis einschlielich aller Unterverzeichnisse und Dateien *)
procedure DeleteDirectory (const Base,Dir           : string;
                           var DCount,FCount,ECount : cardinal); overload;
function DeleteDirectory (const Directory : string) : boolean; overload;


implementation

uses FtGlobal, WinApiUtils, StrUtils;

{ ---------------------------------------------------------------- }
// get time (UTC) of last file write
function GetFileLastWriteTime(const FileName: string): TFileTime;
var
  Handle   : THandle;
  FindData : TWin32FindData;
begin
  Handle := FindFirstFile(PChar(FileName), FindData);
  if Handle <> INVALID_HANDLE_VALUE then begin
    Windows.FindClose(Handle);
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
      Result:=FindData.ftLastWriteTime; Exit;
      end;
    end;
  with Result do begin
    dwLowDateTime:=0; dwHighDateTime:=0;
    end;
  end;

// set time (UTC) of last file write
// CheckTime = true: Change FileTime to actual time if out of range
function SetFileLastWriteTime(const FileName: string; FileTime : TFileTime;
                              CheckTime : boolean) : integer;
var
  Handle   : THandle;
begin
  if CheckTime then begin
    if FileTimeToDateTime(FileTime)>Now+1 then FileTime:=DateTimeToFileTime(Now);
    end;
  Handle:=FileOpen(FileName, fmOpenWrite);
  if Handle=THandle(-1) then Result:=GetLastError
  else begin
    if SetFileTime(Handle,nil,nil,@FileTime) then Result:=0
    else Result:=GetLastError;
    FileClose(Handle);
    end;
  end;

{ ---------------------------------------------------------------- }
// get time (UTC) of last file write
function GetFileLastWriteDateTime(const FileName: string): TDateTime;
var
  Handle   : THandle;
  FindData : TWin32FindData;
  st       : TSystemTime;
  dt       : TDateTime;
begin
  Result:=0;
  Handle := FindFirstFile(PChar(FileName), FindData);
  if Handle <> INVALID_HANDLE_VALUE then begin
    Windows.FindClose(Handle);
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0 then begin
      FileTimeToSystemTime(FindData.ftLastWriteTime,st);
      with st do begin
        if TryEncodeDate(wYear,wMonth,wDay,Result) and
          TryEncodeTime(wHour,wMinute,wSecond,wMilliseconds,dt) then Result:=Result+dt;
        end;
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
// get file or directory timestamps (UTC) from FindData
function GetTimestampsFromFindData(const FindData : TWin32FindData) : TFileTimestamps;
begin
  with Result do begin
    CreationTime:=FindData.ftCreationTime;
    LastAccessTime:=FindData.ftLastAccessTime;
    LastWriteTime:=FindData.ftLastWriteTime;
    Valid:=true;
    end;
  end;

// get file or directory timestamps (UTC)
function GetTimestamps(const AName: string; Directory : boolean) : TFileTimestamps;
var
  Handle   : THandle;
  FindData : TWin32FindData;
  m        : dword;
begin
  Handle := FindFirstFile(PChar(AName), FindData);
  if Handle <> INVALID_HANDLE_VALUE then begin
    Windows.FindClose(Handle);
    if Directory then m:=FILE_ATTRIBUTE_DIRECTORY else m:=0;
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = m then begin
      with Result do begin
        CreationTime:=FindData.ftCreationTime;
        LastAccessTime:=FindData.ftLastAccessTime;
        LastWriteTime:=FindData.ftLastWriteTime;
        Valid:=true;
        Exit;
        end;
      end;
    end;
  FillChar(Result,sizeof(TFileTimestamps),0); // error
  Result.Valid:=false;
  end;

// get file timestamps (UTC)
function GetFileTimestamps(const FileName: string): TFileTimestamps;
begin
  Result:=GetTimestamps(FileName,false);
  end;

// get directory timestamps (UTC)
function GetDirTimestamps(const DirName: string): TFileTimestamps;
begin
  Result:=GetTimestamps(DirName,true);
  end;

// set file or directory timestamps (UTC)
// CheckTime = true: Change FileTime to actual time if out of range
// SetCreationTime = true: Copy timestamp ftCreationTime
function SetFileTimestamps (const FileName: string; Timestamps : TFileTimestamps;
                            CheckTime,SetCreationTime : boolean) : integer;
var
  Handle   : THandle;
  tm       : TFiletime;
  ok       : boolean;
begin
  tm:=DateTimeToFileTime(Now);
  with Timestamps do if Valid then begin
    if CheckTime then begin
      if FileTimeToDateTime(CreationTime)>Now+1 then CreationTime:=tm;
      if FileTimeToDateTime(LastAccessTime)>Now+1 then LastAccessTime:=tm;
      if FileTimeToDateTime(LastWriteTime)>Now+1 then LastWriteTime:=tm;
      end;
    end
  else begin
    CreationTime:=tm;
    LastAccessTime:=tm;
    LastWriteTime:=tm;
    end;
//  Handle:=CreateFile(PChar(FileName),GENERIC_WRITE,0,nil, OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
  Handle:=CreateFile(PChar(FileName),FILE_WRITE_ATTRIBUTES,0,nil, OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
//  Handle:=FileOpen(FileName,fmOpenWrite);
  if Handle=THandle(-1) then Result:=GetLastError
  else with Timestamps do begin
    if SetCreationTime then ok:=SetFileTime(Handle,@CreationTime,nil,@LastWriteTime)
    else ok:=SetFileTime(Handle,nil,nil,@LastWriteTime);
    if ok then Result:=0 else Result:=GetLastError;
    FileClose(Handle);
    end;
  end;

{ ------------------------------------------------------------------- }
// convert Filetime to Delphi time (TDateTime)
function FileTimeToDateTime (ft : TFileTime) : TDateTime;
var
  st : TSystemTime;
  dt : TDateTime;
begin
  FileTimeToSystemTime(ft,st);
  with st do begin
    if TryEncodeDate(wYear,wMonth,wDay,Result) then begin
      if TryEncodeTime(wHour,wMinute,wSecond,wMilliseconds,dt) then Result:=Result+dt;
      end
    else Result:=EncodeDate(1980,1,1);
    end;
  end;

// convert Delphi time (TDateTime) to Filetime
function DateTimeToFileTime (dt : TDateTime) : TFileTime;
var
  st : TSystemTime;
begin
  with st do begin
    DecodeDate(dt,wYear,wMonth,wDay);
    DecodeTime(dt,wHour,wMinute,wSecond,wMilliseconds);
    end;
  SystemTimeToFileTime(st,Result);
  end;

function DateTimeFromFileDate(FileDate : integer) : TDateTime;
begin
  try
    Result:=FileDateToDateTime(FileDate);
  except
    on EConvertError do Result:=Now;
    end;
  end;

{ ------------------------------------------------------------------- }
// convert Delphi time to Unix time (= seconds sinvce 00:00:00 UTC, 1.1.1970) and reverse
function DateTimeToUnixTime (dt : TDateTime) : cardinal;
begin
  Result:=round(SecsPerDay*(dt-25569));
  end;

function UnixTimeTodateTime (ut : cardinal) : TDateTime;
begin
  Result:=ut/SecsPerDay+25569;
  end;

// convert Filetime to Unix time and reverse
function FileTimeToUnixTime (ft : TFileTime) : cardinal;
begin
  Result:=DateTimeToUnixTime (FileTimeToDateTime(ft));
  end;

function UnixTimeToFileTime (ut : cardinal) : TFileTime;
begin
  Result:=DateTimeToFileTime(UnixTimeTodateTime(ut));
  end;

// convert Filetime to seconds
function FileTimeToSeconds (ft : TFileTime) : cardinal;
begin
  Result:=TInt64(ft).AsInt64 div 10000000;
  end;

// convert Filetime to Unix time and reverse
function FileAgeToUnixTime (Age : cardinal) : cardinal;
var
  lft,ft : TFileTime;
begin
  if DosDateTimeToFileTime(LongRec(Age).Hi,LongRec(Age).Lo,lft)
    and LocalFileTimeToFileTime(lft,ft) then
      Result:=FileTimeToUnixTime(ft)
  else Result:=0;
  end;

{ ------------------------------------------------------------------- }
// read file infos (WIN32_FIND_DATA)
function GetFileData(FileName : string; var FileData : TWin32FindData) : boolean;
var
  Handle: THandle;
begin
  Handle:=FindFirstFile(PChar(FileName),FileData);
  Result:=Handle<>INVALID_HANDLE_VALUE;
  if Result then begin
    Windows.FindClose(Handle);
    Result:=(FileData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) = 0;
    end;
  end;

// get DOS file age from file data
function GetFileAge (const FileData: TWin32FindData) : integer;
var
  LocalFileTime: TFileTime;
begin
  FileTimeToLocalFileTime(FileData.ftLastWriteTime,LocalFileTime);
  if not FileTimeToDosDateTime(LocalFileTime,LongRec(Result).Hi,LongRec(Result).Lo) then
    Result:=-1;
  end;

{ ------------------------------------------------------------------- }
// get DOS time stamp of directory (similar to FileAge)
function DirectoryAge(const FileName: string): Integer;
var
  Handle: THandle;
  FindData: TWin32FindData;
  LocalFileTime: TFileTime;
begin
  Handle := FindFirstFile(PChar(FileName), FindData);
  if Handle <> INVALID_HANDLE_VALUE then begin
    Windows.FindClose(Handle);
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY) <> 0 then begin
      FileTimeToLocalFileTime(FindData.ftLastWriteTime, LocalFileTime);
      if FileTimeToDosDateTime(LocalFileTime, LongRec(Result).Hi,
        LongRec(Result).Lo) then Exit;
      end;
    end;
  Result := -1;
  end;

// Set DOS time stamp of directory
function SetDirectoryAge(const FileName: string; Age: Integer): Integer;
begin
  Result:=SetFileAge(FileName,Age);
  end;

function SetFileAge(const FileName: string; Age: Integer): Integer;
var
  f: THandle;
  LocalFileTime, FileTime: TFileTime;
begin
  f:=CreateFile(PChar(FileName),FILE_WRITE_ATTRIBUTES,0,nil, OPEN_EXISTING,FILE_FLAG_BACKUP_SEMANTICS,0);
  if f=THandle(-1) then Result:=GetLastError
  else begin
    if DosDateTimeToFileTime(LongRec(Age).Hi,LongRec(Age).Lo,LocalFileTime) and
      LocalFileTimeToFileTime(LocalFileTime,FileTime) and
      SetFileTime(f, nil, nil, @FileTime) then Result := 0
    else Result := GetLastError;
    FileClose(f);
    end;
  end;

{ ------------------------------------------------------------------- }
// Delete all matching files in directory
procedure DeleteMatchingFiles(const APath,AMask : string);
var
  DirInfo    : TSearchRec;
  Findresult : integer;
begin
  FindResult:=FindFirst (APath+AMask,faAnyFile,DirInfo);
  while (FindResult=0) do begin
    EraseFile(APath+DirInfo.Name);
    FindResult:=FindNext(DirInfo)
    end;
  FindClose(DirInfo);
  end;

{ ---------------------------------------------------------------- }
(* Diese Funktion prft, ob eine Text-Datei vorhanden ist und ffnet sie *)
function ExistsFile (var f : TextFile) : boolean;
begin
  (*$i-*) reset (f) (*$i+*);
  Result:=ioresult=0;
  end;

//-----------------------------------------------------------------------------
// Prfen, ob in einem Verzeichnis eine Datei existiert, die der Maske entspricht
function FileMatchesMask (Dir,Mask : string) : string;
var
  FInfo      : TSearchRec;
begin
  Result:='';
  if FindFirst (IncludeTrailingPathDelimiter(Dir)+Mask,faArchive,FInfo)=0 then
    Result:=IncludeTrailingPathDelimiter(Dir)+FInfo.Name;
  Sysutils.FindClose (FInfo);
  end;

{ ------------------------------------------------------------------- }
// generate error message
constructor ECopyError.Create (const ErrString : string;
                               ErrNR           : integer);
begin
  inherited Create (ErrString);
  CopyError:=ErrNr;
  end;

{ ------------------------------------------------------------------- }
// copy file with attribute and timestamp - uses CopyFile from WindowsAPI 
procedure FileCopy (srcfilename,destfilename : String);
var
  ok : boolean;
begin
  if FileExists (srcfilename) and (length(destfilename)>0) then begin
    if FileExists(destfilename) then ok:=FileSetAttr(destfilename,0)=0
    else ok:=true;
    if ok then begin
      if not CopyFile(pchar(srcfilename),pchar(destfilename),false) then begin
        raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
        end;
      end
    else begin
      raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
      end;
    end;
  end;

// copy file with timestamp and set new attribute
procedure FileCopy (srcfilename,destfilename : String; Attr : word);
begin
  try
    FileCopy(srcfilename,destfilename);
    FileSetAttr(destfilename,Attr);
  except
    raise EInOutError.Create (Format(rsErrCopy,[destfilename]));
    end;
  end;

{ ------------------------------------------------------------------- }
// Copy file with timestamp
procedure CopyFileTS (const srcfilename,destfilename : String; BlockSize : integer);
var
  srcfile, destfile : TFileStream;
  FTime             : TFileTime;
  Buffer            : pointer;
  NRead,NWrite      : Integer;
  Attr              : word;
begin
  if ExistingFile (srcfilename) and (length(destfilename)>0) then begin
    GetMem(Buffer,BlockSize);
    try
      FTime:=GetFileLastWriteTime(srcfilename);
      Attr:=FileGetAttr(srcfilename);
      try
        srcfile:=TFileStream.Create(srcfilename,fmOpenRead+fmShareDenyNone);
      except
        on EFOpenError do
          raise ECopyError.Create (Format(rsErrOpening,[srcfilename]),errFileOpen);
        end;
      // Ziel immer berschreiben
      if ExistingFile(destfilename) then begin
        if FileSetAttr(destfilename,faArchive)<>0 then begin
          try srcfile.Free; except end;
          raise ECopyError.Create (Format(rsErrCreating,[destfilename]),errFileCreate);
          end;
        end;
      try
        destfile:=TFileStream.Create(destfilename,fmCreate);
      except
        on EFCreateError do begin
          try srcfile.Free; except end;
          raise ECopyError.Create (Format(rsErrCreating,[destfilename]),errFileCreate);
          end;
        end;
      repeat
        try
          NRead:=srcfile.Read(Buffer^,BlockSize);
        except
          on EReadError do
            raise ECopyError.Create (Format(rsErrReading,[srcfilename]),errFileRead);
          end;
        try
          NWrite:=destfile.Write(Buffer^,NRead);
          if NWrite<NRead then  // Ziel-Medium voll
            raise ECopyError.Create (Format(rsErrWriting,[destfilename]),errFileWrite);
        except
          on EWriteError do
            raise ECopyError.Create (Format(rsErrWriting,[destfilename]),errFileWrite);
          end;
        until NRead<BlockSize;
      if  destfile.Size<>srcfile.Size then begin
      // z.B. wenn "srcfile" gelockt ist (siehe LockFile)
        srcfile.Free; destfile.Free;
        raise ECopyError.Create (Format(rsErrReading,[srcfilename]),errFileRead);
        end;
      try
        srcfile.Free;
      except
        on EFileStreamError do
          raise ECopyError.Create (Format(rsErrClosing,[srcfilename]),errFileClose);
        end;
      try
        destfile.Free;
      except
        on EFileStreamError do
          raise ECopyError.Create (Format(rsErrClosing,[destfilename]),errFileClose);
        end;
      if SetFileLastWriteTime(destfilename,FTime,true)=0 then begin
        if FileSetAttr(destfilename,Attr)>0 then
          raise ECopyError.Create (Format(rsErrSetAttr,[destfilename]),errWarning+errFileAttr);
        end
      else
        raise ECopyError.Create (Format(rsErrTimeStamp,[destfilename]),errWarning+errFileTS);
    finally
      FreeMem(Buffer,BlockSize);
      end;
    end;
  end;

{ ---------------------------------------------------------------- }
// convert file attribute to string
function FileAttrToString(Attr : word) : string;

var
  s : string;
begin
  s:='';
  if Attr and faReadOnly =0 then s:=s+'-' else s:=s+'r';
  if Attr and faArchive =0 then s:=s+'-' else s:=s+'a';
  if Attr and faHidden =0 then s:=s+'-' else s:=s+'h';
  if Attr and faSysFile =0 then s:=s+'-' else s:=s+'s';
//  if Attr and faSymLink	=0 then s:=s+'-' else s:=s+'L';
  Result:=s;
  end;

{ ---------------------------------------------------------------- }
function GetFileSize (const FileData: TWin32FindData) : int64;
begin
  with TInt64(Result) do begin
    Lo:=FileData.nFileSizeLow; Hi:=FileData.nFileSizeHigh;
    end;
  end;

function LongFileSize (const FileName : string) : int64;
var
  FileData : TWin32FindData;
begin
  if GetFileData(FileName,FileData) then Result:=GetFileSize(FileData)
  else Result:=0;
  end;

const
  SeekStep = $80000000;

{ ------------------------------------------------------------------- }
// Convert SearchRec to FileInfo
function SearchRecToFileInfo (SearchRec : TSearchRec) : TFileInfo;
begin
  with result do begin
    Name:=SearchRec.Name;
    Attr:=SearchRec.Attr;
    Size:=GetFileSize(SearchRec.FindData);
    DateTime:=DateTimeFromFileDate(SearchRec.Time);
    end;
  end;

{ ------------------------------------------------------------------- }
// search abs. Position in stream, even if >= 2^32
procedure SeekPosition (sFile : TFileStream; Pos : int64; Origin : word);
var
  n : int64;
begin
  sFile.Seek(0,Origin);  //seek does not support int64
  repeat
    if Pos>=SeekStep then n:=SeekStep-1 else n:=Pos;
    dec(Pos,n);
    sFile.Seek(n,soFromCurrent);
    until Pos<=0;
  end;

{ ------------------------------------------------------------------- }
// check if file exists
// replaces function FileExists in unit SysUtils
// reason: FileExists returns an error on existing files with illegale date
function ExistingFile (const FileName: string): boolean;
var
  Handle: THandle;
  FindData: TWin32FindData;
begin
  Handle:=FindFirstFile(PChar(FileName),FindData);
  if Handle<>INVALID_HANDLE_VALUE then begin
    Windows.FindClose(Handle);
    Result:=true;
    end
  else Result:=false;
  end;

{ ------------------------------------------------------------------- }
// Delete file even if readonly
function EraseFile(const FileName: string): Boolean;
begin
  if ExistingFile(FileName) then begin
    Result:=FileSetAttr(FileName,faArchive)=0;
    if Result then Result:=DeleteFile(FileName);
    end
  else Result:=false;
  end;

{ ------------------------------------------------------------------- }
(* kein spezielles Verzeichnis (selbst und eins zurck) *)
function NotSpecialDir (const Name : string; StartChars : array of char) : boolean;
var
  i : integer;
begin
  Result:=(Name<>'.') and (Name<>'..');
  if Result and (length(StartChars)>0) then begin
    for i:=Low(StartChars) to High(StartChars) do Result:=Result and (Pos(StartChars[i],Name)<>1);
    end;
  end;

function NotSpecialDir (const Name : string) : boolean;
begin
  Result:=NotSpecialDir(Name,[]);
  end;

function IsRecursiveParsePoint (const Path : string; Attr : integer; AllowJunctions : boolean) : boolean;
var
  s : string;
begin
  Result:=false;
  if Attr=-1 then Exit;
  if (Attr and FILE_ATTRIBUTE_REPARSE_POINT <>0) then begin  //directory entry is a reparse point
    if AllowJunctions then begin  // check if recursive link
      s:=GetReparseMountPoint(Path);
      if (length(s)>0) and AnsiStartsText(s,Path) then Result:=true;  // avoid recursions
      end
    else Result:=true;
    end;
  end;

function IsRecursiveParsePoint (const Path : string; AllowJunctions : boolean) : boolean;
begin
  Result:=IsRecursiveParsePoint (Path,FileGetAttr(Path),AllowJunctions);
  end;

{ ------------------------------------------------------------------- }
// Check if directory is empty
function IsEmptyDir (const Directory : string) : boolean;
var
  DirInfo    : TSearchRec;
  FindResult,
  n          : integer;
begin
  n:=0;
  FindResult:=FindFirst (IncludeTrailingPathDelimiter(Directory)+'*.*',faAnyFile,DirInfo);
  while (FindResult=0) do with DirInfo do begin
    if NotSpecialDir(Name) then inc(n);
    FindResult:=FindNext (DirInfo);
    end;
  FindClose(DirInfo);
  Result:=n=0;
  end;

{ ---------------------------------------------------------------- }
// Dateien in einem Verzeichnis zhlen und Gesamtgre bestimmen
procedure CountFiles (const Base,Dir : string; IncludeSubDir : boolean;
                      var FileCount : cardinal; var FileSize : int64);
var
  DirInfo    : TSearchRec;
  Findresult : integer;
  sd         : string;
begin
  sd:=IncludeTrailingPathDelimiter(IncludeTrailingPathDelimiter(Base)+Dir);
  FindResult:=FindFirst (sd+'*.*',faAnyFile,DirInfo);
  while FindResult=0 do with DirInfo do begin
    if IncludeSubDir and NotSpecialDir(Name) and ((Attr and faDirectory)<>0) then
      CountFiles (Base,IncludeTrailingPathDelimiter(Dir)+DirInfo.Name,IncludeSubDir,FileCount,FileSize);
    FindResult:=FindNext(DirInfo);
    end;
  FindClose(DirInfo);
  FindResult:=FindFirst (sd+'*.*',faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
  while FindResult=0 do with DirInfo do begin
    if NotSpecialDir(Name) then begin
      inc(FileCount); inc(FileSize,GetFileSize(DirInfo.FindData));
      end;
    FindResult:=FindNext(DirInfo);
    end;
  FindClose(DirInfo);
  end;

function DirFiles (const Directory : string; IncludeSubDir : boolean) : cardinal;
var
  n : int64;
begin
  Result:=0; n:=0;
  CountFiles(Directory,'',IncludeSubDir,Result,n);
  end;

function DirSize (const Directory : string; IncludeSubDir : boolean) : int64;
var
  n : cardinal;
begin
  Result:=0; n:=0;
  CountFiles(Directory,'',IncludeSubDir,n,Result);
  end;

(* Lsche ein Verzeichnis einschlielich aller Unterverzeichnisse und Dateien *)
procedure DeleteDirectory (const Base,Dir           : string;
                           var DCount,FCount,ECount : cardinal);
var
  DirInfo    : TSearchRec;
  fc,dc,
  Findresult : integer;
  s,sd       : string;
begin
  if length(Dir)>0 then sd:=IncludeTrailingPathDelimiter(Base)+Dir else sd:=Base;
  if DirectoryExists(sd) then begin
    FindResult:=FindFirst (IncludeTrailingPathDelimiter(sd)+'*.*',faAnyFile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) and ((Attr and faDirectory)<>0) then
        DeleteDirectory(Base,IncludeTrailingPathDelimiter(Dir)+DirInfo.Name,DCount,FCount,ECount);
      FindResult:=FindNext(DirInfo);
      end;
    FindClose(DirInfo);
    fc:=0; dc:=0;
    FindResult:=FindFirst (IncludeTrailingPathDelimiter(sd)+'*.*',faArchive+faReadOnly+faHidden+faSysfile,DirInfo);
    while FindResult=0 do with DirInfo do begin
      if NotSpecialDir(Name) then begin
        inc(fc);
        (* Dateien lschen *)
        s:=IncludeTrailingPathDelimiter(sd)+Name;
        FileSetAttr(s,faArchive);
        if DeleteFile(s) then begin
          inc(FCount); inc(dc);
          end
        else inc(ECount);  // Fehler
        end;
      FindResult:=FindNext(DirInfo);
      end;
    FindClose(DirInfo);
    if (fc=dc) then begin   // Verzeichnis leer ==> lschen
      FileSetAttr(sd,0);    // Attribute zum Lschen entfernen
      if RemoveDir(sd) then inc(DCount) else inc(ECount);
      end;
    end;
  end;

function DeleteDirectory (const Directory : string) : boolean;
var
  fc,dc,ec : cardinal;
begin
  fc:=0; dc:=0; ec:=0;
  DeleteDirectory(Directory,'',dc,fc,ec);
  Result:=ec=0;
  end;

end.
